import asyncio
import signal
from pathlib import Path
from socket import gethostname
from typing import Dict, List

import structlog

from indexify.proto.executor_api_pb2 import ExecutorStatus

from .blob_store.blob_store import BLOBStore
from .channel_manager import ChannelManager
from .function_allowlist import (
    FunctionURI,
    function_allowlist_to_indexed_dict,
    parse_function_uris,
)
from .function_executor.server.function_executor_server_factory import (
    FunctionExecutorServerFactory,
)
from .host_resources.host_resources import HostResourcesProvider
from .metrics.executor import (
    metric_executor_info,
    metric_executor_state,
)
from .monitoring.desired_state_handler import DesiredStateHandler
from .monitoring.health_check_handler import HealthCheckHandler
from .monitoring.health_checker.health_checker import HealthChecker
from .monitoring.prometheus_metrics_handler import PrometheusMetricsHandler
from .monitoring.reported_state_handler import ReportedStateHandler
from .monitoring.server import MonitoringServer
from .monitoring.startup_probe_handler import StartupProbeHandler
from .state_reconciler import ExecutorStateReconciler
from .state_reporter import ExecutorStateReporter

metric_executor_state.state("starting")


class Executor:
    def __init__(
        self,
        id: str,
        version: str,
        labels: Dict[str, str],
        cache_path: Path,
        health_checker: HealthChecker,
        function_uris: List[str],
        function_executor_server_factory: FunctionExecutorServerFactory,
        server_addr: str,
        grpc_server_addr: str,
        config_path: str | None,
        monitoring_server_host: str,
        monitoring_server_port: int,
        blob_store: BLOBStore,
        host_resources_provider: HostResourcesProvider,
    ):
        self._logger = structlog.get_logger(module=__name__, executor_id=id)
        protocol: str = "http"
        if config_path:
            self._logger.info("running the extractor with TLS enabled")
            protocol = "https"

        self._startup_probe_handler = StartupProbeHandler()
        self._channel_manager = ChannelManager(
            server_address=grpc_server_addr,
            config_path=config_path,
            logger=self._logger,
        )
        function_allowlist: List[FunctionURI] = parse_function_uris(function_uris)
        self._state_reporter = ExecutorStateReporter(
            executor_id=id,
            version=version,
            labels=labels,
            function_allowlist=function_allowlist,
            channel_manager=self._channel_manager,
            host_resources_provider=host_resources_provider,
            health_checker=health_checker,
            logger=self._logger,
        )
        self._state_reporter.update_executor_status(
            ExecutorStatus.EXECUTOR_STATUS_STARTING_UP
        )
        self._state_reconciler = ExecutorStateReconciler(
            executor_id=id,
            function_executor_server_factory=function_executor_server_factory,
            base_url=f"{protocol}://{server_addr}",
            config_path=config_path,
            cache_path=cache_path,
            blob_store=blob_store,
            channel_manager=self._channel_manager,
            state_reporter=self._state_reporter,
            logger=self._logger,
        )
        self._monitoring_server = MonitoringServer(
            host=monitoring_server_host,
            port=monitoring_server_port,
            startup_probe_handler=self._startup_probe_handler,
            health_probe_handler=HealthCheckHandler(health_checker),
            metrics_handler=PrometheusMetricsHandler(),
            reported_state_handler=ReportedStateHandler(self._state_reporter),
            desired_state_handler=DesiredStateHandler(self._state_reconciler),
        )
        self._run_aio_task: asyncio.Task | None = None
        self._shutdown_aio_task: asyncio.Task | None = None

        executor_info: Dict[str, str] = {
            "id": id,
            "version": version,
            "cache_path": str(cache_path),
            "server_addr": server_addr,
            "grpc_server_addr": str(grpc_server_addr),
            "config_path": str(config_path),
            "hostname": gethostname(),
        }
        for key, value in labels.items():
            executor_info["label_" + key] = value
        executor_info.update(function_allowlist_to_indexed_dict(function_allowlist))
        metric_executor_info.info(executor_info)

    def run(self):
        asyncio.new_event_loop()

        self._run_aio_task = asyncio.get_event_loop().create_task(
            self._run(),
            name="executor startup and run loop",
        )

        try:
            asyncio.get_event_loop().run_until_complete(self._run_aio_task)
        except asyncio.CancelledError:
            pass  # Expected exception on shutdown

    async def _run(self):
        for signum in [
            signal.SIGABRT,
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGQUIT,
            signal.SIGHUP,
        ]:
            asyncio.get_event_loop().add_signal_handler(
                signum, self._shutdown_signal_handler, asyncio.get_event_loop()
            )

        asyncio.create_task(
            self._monitoring_server.run(), name="monitoring server runner"
        )
        self._state_reporter.update_executor_status(
            ExecutorStatus.EXECUTOR_STATUS_RUNNING
        )
        self._state_reporter.run()
        self._state_reconciler.run()
        metric_executor_state.state("running")
        self._startup_probe_handler.set_ready()

        # Run the Executor forever until it is shut down.
        while True:
            await asyncio.sleep(10)

    def _shutdown_signal_handler(self, loop):
        if self._shutdown_aio_task is None:
            self._shutdown_aio_task = loop.create_task(
                self._shutdown(), name="executor shutdown"
            )

    async def _shutdown(self):
        self._logger.info("shutting down Executor")
        metric_executor_state.state("shutting_down")

        # Shutdown state reconciler first because it changes reported state on shutdown.
        await self._state_reconciler.shutdown()

        # Do one last state report with STOPPED status. This reduces latency in the system.
        self._state_reporter.update_executor_status(
            ExecutorStatus.EXECUTOR_STATUS_STOPPED
        )
        await self._state_reporter.shutdown()
        await self._channel_manager.destroy()
        await self._monitoring_server.shutdown()
        self._run_aio_task.cancel()



from typing import Any

import grpc

# Timeout for Function Executor Server startup in seconds. The timeout is counted from
# the moment when a server just started.
FUNCTION_EXECUTOR_SERVER_READY_TIMEOUT_SEC = 5


class FunctionExecutorServer:
    """Abstract interface for a Function Executor Server.

    FunctionExecutorServer is a class that executes tasks for a particular function.
    The communication with FunctionExecutorServer is typicall done via gRPC.
    """

    async def create_channel(self, logger: Any) -> grpc.aio.Channel:
        """Creates a new async gRPC channel to the Function Executor Server.

        The channel is in ready state. It can only be used in the same thread where the
        function was called. Caller should close the channel when it's no longer needed.

        Raises Exception if an error occurred."""
        raise NotImplementedError



import os
import traceback

import click
from tensorlake.applications.remote.deploy import deploy_applications


@click.command(
    short_help="Deploys applications defined in <application-file-path> .py file to Indexify"
)
@click.argument(
    "application-file-path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "-u",
    "--upgrade-running-requests",
    is_flag=True,
    default=False,
    help="Upgrade requests that are already queued or running to use the new deployed version of the applications",
)
def deploy(
    application_file_path: str,
    upgrade_running_requests: bool,
):
    click.echo(f"Preparing deployment for applications from {application_file_path}")

    try:
        deploy_applications(
            applications_file_path=application_file_path,
            upgrade_running_requests=upgrade_running_requests,
            load_source_dir_modules=True,
        )
    except Exception as e:
        click.secho(
            f"Applications could not be deployed, please check the error message:",
            fg="red",
        )
        traceback.print_exception(e)
        raise click.Abort

    click.secho(f"Successfully deployed the applications", fg="green")



import asyncio
import os
import signal
import socket
from typing import Any

from .function_executor_server_factory import (
    FunctionExecutorServerConfiguration,
    FunctionExecutorServerFactory,
)
from .subprocess_function_executor_server import SubprocessFunctionExecutorServer


class SubprocessFunctionExecutorServerFactory(FunctionExecutorServerFactory):
    def __init__(self, verbose_logs: bool) -> None:
        super().__init__()
        self._verbose_logs = verbose_logs

    async def create(
        self, config: FunctionExecutorServerConfiguration, logger: Any
    ) -> SubprocessFunctionExecutorServer:
        logger = logger.bind(module=__name__)
        port: int | None = None

        if len(config.secret_names) > 0:
            logger.warning(
                "Subprocess Function Executor does not support secrets. Please supply secrets as environment variables.",
                secret_names=config.secret_names,
            )

        try:
            port = _find_free_localhost_tcp_port()
            args = [
                f"--executor-id={config.executor_id}",  # use = as executor_id can start with -
                f"--function-executor-id={config.function_executor_id}",
                "--address",
                _server_address(port),
            ]
            if self._verbose_logs:
                args.append("--dev")
            # Run the process with our stdout, stderr. We want to see process logs and exceptions in our process output.
            # This is useful for dubugging. Customer function stdout and stderr is captured and returned in the response
            # so we won't see it in our process outputs. This is the right behavior as customer function stdout and stderr
            # contains private customer data.
            proc: asyncio.subprocess.Process = await asyncio.create_subprocess_exec(
                "function-executor",
                *args,
                # TODO: pass `process_group=0` instead of the depricated `preexec_fn` once we only support Python 3.11+.
                preexec_fn=_new_process_group,
            )
            return SubprocessFunctionExecutorServer(
                process=proc,
                port=port,
                address=_server_address(port),
            )
        except Exception as e:
            logger.error(
                "failed starting a new Function Executor process at port {port}",
                exc_info=e,
            )
            raise

    async def destroy(
        self, server: SubprocessFunctionExecutorServer, logger: Any
    ) -> None:
        proc: asyncio.subprocess.Process = server._proc
        port: int = server._port
        logger = logger.bind(
            module=__name__,
            pid=proc.pid,
            port=port,
        )

        try:
            if proc.returncode is not None:
                # The process already exited and was waited() sucessfully.
                return

            if os.name == "posix":
                # On POSIX systems, we can kill the whole process group so processes forked by customer code are also killed.
                # This should be done before proc.kill because PG processes get their own PG when their PG leader dies.
                os.killpg(proc.pid, signal.SIGKILL)
            proc.kill()
            await proc.wait()
        except Exception as e:
            logger.error(
                "failed to cleanup Function Executor process",
                exc_info=e,
            )


# Function Executors are only listening on localhost so external connections to them are not possible.
# This is a security measure. Also Executor <-> Function Executor communication is always local and
# don't support Function Executors running on a different host.
_FUNCTION_EXECUTOR_SERVER_HOSTNAME = "localhost"


def _server_address(port: int) -> str:
    return f"{_FUNCTION_EXECUTOR_SERVER_HOSTNAME}:{port}"


def _new_process_group() -> None:
    """Creates a new process group with ID equal to the current process PID. POSIX only."""
    os.setpgid(0, 0)


def _find_free_localhost_tcp_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind((_FUNCTION_EXECUTOR_SERVER_HOSTNAME, 0))
        _, port = sock.getsockname()
        return port



from dataclasses import dataclass
from typing import Any, List

import psutil

from .nvidia_gpu import NvidiaGPUInfo
from .nvidia_gpu_allocator import NvidiaGPUAllocator


@dataclass
class HostResources:
    cpu_count: int
    memory_mb: int
    disk_mb: int
    gpus: List[NvidiaGPUInfo]


class HostResourcesProvider:
    """
    HostResourcesProvider is a class that provides information about the host resources.
    """

    def __init__(
        self,
        gpu_allocator: NvidiaGPUAllocator,
        function_executors_ephimeral_disks_path: str,
        host_overhead_cpus: int,
        host_overhead_memory_gb: int,
        host_overhead_function_executors_ephimeral_disks_gb: int,
    ):
        """Creates a HostResourcesProvider.

        Args:
            gpu_allocator: The GPU allocator to use for GPU information.
            function_executors_ephimeral_disks_path: The path to file system used as ephimeral disk space by Function Executors.
            host_overhead_cpus: The number of CPUs reserved for use by host (can't be used by Function Executors).
            host_overhead_memory_gb: The amount of memory reserved for use by host (can't be used by Function Executors).
            host_overhead_function_executors_ephimeral_disks_gb: The amount of ephimeral disk space reserved for use by host (can't be used by Function Executors).
        """
        self._gpu_allocator: NvidiaGPUAllocator = gpu_allocator
        self._function_executors_ephimeral_disks_path: str = (
            function_executors_ephimeral_disks_path
        )
        self._host_overhead_cpus: int = host_overhead_cpus
        self._host_overhead_memory_gb: int = host_overhead_memory_gb
        self._host_overhead_function_executors_ephimeral_disks_gb: int = (
            host_overhead_function_executors_ephimeral_disks_gb
        )

    def total_host_resources(self, logger: Any) -> HostResources:
        """Returns all hardware resources that exist at the host.

        Raises Exception on error.
        """
        logger = logger.bind(module=__name__)

        # If users disable Hyper-Threading in OS then we'd only see physical cores here.
        # This allows users to control if logical or physical cores are used for resource
        # reporting and for running the functions.
        cpu_count: int | None = psutil.cpu_count(logical=True)
        if cpu_count is None:
            logger.warning(
                "Unable to determine CPU count. Defaulting to 0.",
                cpu_count=cpu_count,
            )
            cpu_count = 0

        memory_mb: int = int(psutil.virtual_memory().total / 1024 / 1024)
        disk_mb = int(
            psutil.disk_usage(self._function_executors_ephimeral_disks_path).total
            / 1024
            / 1024
        )
        all_gpus: List[NvidiaGPUInfo] = self._gpu_allocator.list_all()

        return HostResources(
            cpu_count=cpu_count,
            memory_mb=memory_mb,
            disk_mb=disk_mb,
            gpus=all_gpus,
        )

    def total_function_executor_resources(self, logger: Any) -> HostResources:
        """Returns all hardware resources on the host that are usable by Function Executors.

        Raises Exception on error.
        """
        total_resources: HostResources = self.total_host_resources(logger=logger)
        return HostResources(
            cpu_count=max(0, total_resources.cpu_count - self._host_overhead_cpus),
            memory_mb=max(
                0, total_resources.memory_mb - self._host_overhead_memory_gb * 1024
            ),
            disk_mb=max(
                0,
                total_resources.disk_mb
                - self._host_overhead_function_executors_ephimeral_disks_gb * 1024,
            ),
            gpus=total_resources.gpus,
        )



import asyncio
from dataclasses import dataclass

from indexify.proto.executor_api_pb2 import Allocation

from .allocation_input import AllocationInput
from .allocation_output import AllocationOutput


@dataclass
class AllocationInfo:
    """Object used to track a allocation during its full lifecycle in the FunctionExecutorController."""

    allocation: Allocation
    # Timeout in ms for running customer code.
    allocation_timeout_ms: int
    # time.monotonic() timestamp
    start_time: float
    # time.monotonic() timestamp when the task was prepared for execution
    prepared_time: float = 0.0
    # True if the task was cancelled.
    is_cancelled: bool = False
    # aio task that is currently executing a lifecycle step of this task.
    aio_task: asyncio.Task | None = None
    # Input of the function if task allocation was prepared successfully.
    input: AllocationInput | None = None
    # Output of the task allocation, always set when the allocation is completed.
    output: AllocationOutput | None = None
    # True if the task allocation is fully completed and was added to state reporter.
    is_completed: bool = False



import asyncio
import os
from pathlib import Path
from typing import Any

import nanoid
from tensorlake.function_executor.proto.function_executor_pb2 import (
    SerializedObject,
    SerializedObjectEncoding,
    SerializedObjectManifest,
)

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.proto.executor_api_pb2 import (
    DataPayload,
    DataPayloadEncoding,
    FunctionExecutorDescription,
)

from .metrics.downloads import (
    metric_application_download_errors,
    metric_application_download_latency,
    metric_application_downloads,
    metric_applications_from_cache,
)


async def download_application_code(
    function_executor_description: FunctionExecutorDescription,
    cache_path: Path,
    blob_store: BLOBStore,
    logger: Any,
) -> SerializedObject:
    logger = logger.bind(module=__name__)
    with (
        metric_application_download_errors.count_exceptions(),
        metric_application_download_latency.time(),
    ):
        metric_application_downloads.inc()
        return await _download_application_code(
            function_executor_description=function_executor_description,
            cache_path=cache_path,
            blob_store=blob_store,
            logger=logger,
        )


async def _download_application_code(
    function_executor_description: FunctionExecutorDescription,
    cache_path: Path,
    blob_store: BLOBStore,
    logger: Any,
) -> SerializedObject:
    # Cache application to reduce load on the server.
    application_path = os.path.join(
        str(cache_path),
        "application_cache",
        function_executor_description.function.namespace,
        function_executor_description.function.application_name,
        function_executor_description.function.application_version,
    )
    # Filesystem operations are synchronous.
    # Run in a separate thread to not block the main event loop.
    application: SerializedObject | None = await asyncio.to_thread(
        _read_cached_application, path=application_path
    )
    if application is not None:
        metric_applications_from_cache.inc()
        return application

    data: bytes = await blob_store.get(
        uri=function_executor_description.application.uri, logger=logger
    )
    application: SerializedObject = SerializedObject(
        manifest=serialized_object_manifest_from_data_payload_proto(
            function_executor_description.application
        ),
        data=data,
    )

    # Filesystem operations are synchronous.
    # Run in a separate thread to not block the main event loop.
    # We don't need to wait for the write completion so we use create_task.
    asyncio.create_task(
        asyncio.to_thread(
            _write_cached_application,
            path=application_path,
            application=application,
            cache_path=cache_path,
        ),
        name="application cache write",
    )

    return application


def _read_cached_application(path: str) -> SerializedObject | None:
    if not os.path.exists(path):
        return None

    with open(path, "rb") as f:
        return SerializedObject.FromString(f.read())


def _write_cached_application(
    path: str, application: SerializedObject, cache_path: Path
) -> None:
    if os.path.exists(path):
        # Another allocation already cached the application.
        return None

    tmp_path = os.path.join(
        str(cache_path), "task_application_cache", nanoid.generate()
    )
    os.makedirs(os.path.dirname(tmp_path), exist_ok=True)
    with open(tmp_path, "wb") as f:
        f.write(application.SerializeToString())
    os.makedirs(os.path.dirname(path), exist_ok=True)
    # Atomically rename the fully written file at tmp path.
    # This allows us to not use any locking because file link/unlink
    # are atomic operations at filesystem level.
    # This also allows to share the same cache between multiple Executors.
    os.replace(tmp_path, path)


def serialized_object_manifest_from_data_payload_proto(
    data_payload: DataPayload,
) -> SerializedObjectManifest:
    """Converts the given data payload into SerializedObjectManifest accepted by Function Executor.

    Raises ValueError if the supplied data payload can't be converted.
    """
    so_manifest: SerializedObjectManifest = SerializedObjectManifest(
        # Server currently ignores encoding version so we set it to default 0.
        encoding_version=(
            data_payload.encoding_version
            if data_payload.HasField("encoding_version")
            else 0
        ),
        sha256_hash=data_payload.sha256_hash,
        size=data_payload.size,
        metadata_size=data_payload.metadata_size,
    )
    if data_payload.HasField("content_type"):
        so_manifest.content_type = data_payload.content_type
    if data_payload.HasField("source_function_call_id"):
        so_manifest.source_function_call_id = data_payload.source_function_call_id
    # data_payload.id is not used.

    if data_payload.encoding == DataPayloadEncoding.DATA_PAYLOAD_ENCODING_BINARY_PICKLE:
        so_manifest.encoding = (
            SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_BINARY_PICKLE
        )
    elif data_payload.encoding == DataPayloadEncoding.DATA_PAYLOAD_ENCODING_UTF8_TEXT:
        so_manifest.encoding = (
            SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_UTF8_TEXT
        )
    elif data_payload.encoding == DataPayloadEncoding.DATA_PAYLOAD_ENCODING_UTF8_JSON:
        so_manifest.encoding = (
            SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_UTF8_JSON
        )
    elif data_payload.encoding == DataPayloadEncoding.DATA_PAYLOAD_ENCODING_BINARY_ZIP:
        so_manifest.encoding = (
            SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_BINARY_ZIP
        )
    elif data_payload.encoding == DataPayloadEncoding.DATA_PAYLOAD_ENCODING_RAW:
        so_manifest.encoding = SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_RAW
    else:
        raise ValueError(
            f"Can't convert data payload {data_payload} into serialized object"
        )

    return so_manifest



class HealthCheckResult:
    def __init__(self, checker_name: str, is_success: bool, status_message: str):
        self.checker_name: str = checker_name
        self.is_success: bool = is_success
        self.status_message: str = status_message


class HealthChecker:
    """Abstract base class for health checkers."""

    def server_connection_state_changed(self, is_healthy: bool, status_message: str):
        """Handle changes in server connection state."""
        raise NotImplementedError("Subclasses must implement this method.")

    async def check(self) -> HealthCheckResult:
        raise NotImplementedError("Subclasses must implement this method.")



import os
from typing import Any, List

from .events import BaseEvent

_DEBUG_EVENT_LOOP: bool = (
    os.getenv("INDEXIFY_FUNCTION_EXECUTOR_CONTROLLER_DEBUG_EVENT_LOOP", "0")
) == "1"


def debug_print_processing_event(event: BaseEvent, logger: Any) -> None:
    if _DEBUG_EVENT_LOOP:
        logger.debug(
            "processing event in control loop",
            fe_event=str(event),
        )


def debug_print_adding_event(event: BaseEvent, source: str, logger: Any) -> None:
    if _DEBUG_EVENT_LOOP:
        logger.debug(
            "adding event to control loop",
            source=source,
            fe_event=str(event),
        )


def debug_print_events(events: List[BaseEvent], logger: Any) -> None:
    if _DEBUG_EVENT_LOOP:
        if len(events) == 0:
            logger.debug("no events n control loop")
        else:
            logger.debug(
                "events in control loop",
                count=len(events),
                fe_events=[str(event) for event in events],
            )



import asyncio
from typing import Any


async def shielded_await(aio_task: asyncio.Task, logger: Any) -> Any:
    """Awaits the supplied task and ignores cancellations until it's done.

    Cancels itself if the task is cancelled once the task is done.
    """
    cancelled_error: asyncio.CancelledError | None = None

    while not aio_task.done():
        try:
            # Shield to make sure that task is not cancelled.
            await asyncio.shield(aio_task)
        except asyncio.CancelledError as e:
            logger.info(
                "ignoring aio task cancellation until it's finished",
                task_name=aio_task.get_name(),
            )
            cancelled_error = e

    if cancelled_error is not None:
        raise cancelled_error
    else:
        return aio_task.result()



import asyncio
from typing import Any, Optional

import boto3
from botocore.config import Config as BotoConfig
from botocore.exceptions import ClientError as BotoClientError

_MAX_RETRIES = 3


class S3BLOBStore:
    def __init__(self):
        self._s3_client: Optional[Any] = None

    def _lazy_create_client(self):
        """Creates S3 client if it doesn't exist.

        We create the client lazily only if S3 is used.
        This is because S3 BLOB store is always created by Executor
        and the creation will fail if user didn't configure S3 credentials and etc.
        """
        if self._s3_client is not None:
            return

        # The credentials and etc are fetched by boto3 library automatically following
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#configuring-credentials
        # This provides a lot of flexibility for the user and follows a well-known and documented logic.
        self._s3_client = boto3.client(
            "s3",
            config=BotoConfig(
                # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html#standard-retry-mode
                retries={
                    "max_attempts": _MAX_RETRIES,
                    "mode": "standard",
                }
            ),
        )

    async def get(self, uri: str, logger: Any) -> bytes:
        """Returns binary value stored in S3 object at the supplied URI.

        The URI must be S3 URI (starts with "s3://").
        Raises Exception on error. Raises KeyError if the object doesn't exist.
        """
        try:
            self._lazy_create_client()
            bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
            response = await asyncio.to_thread(
                self._s3_client.get_object, Bucket=bucket_name, Key=key
            )
            return response["Body"].read()
        except BotoClientError as e:
            logger.error("failed to get S3 object", uri=uri, exc_info=e)

            if e.response["Error"]["Code"] == "NoSuchKey":
                raise KeyError(f"Object {key} does not exist in bucket {bucket_name}")
            raise
        except Exception as e:
            logger.error("failed to get S3 object", uri=uri, exc_info=e)
            raise

    async def presign_get_uri(self, uri: str, expires_in_sec: int, logger: Any) -> str:
        """Returns a presigned URI for getting the S3 object at the supplied URI."""
        self._lazy_create_client()
        bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
        try:
            s3_uri: str = await asyncio.to_thread(
                self._s3_client.generate_presigned_url,
                ClientMethod="get_object",
                Params={"Bucket": bucket_name, "Key": key},
                ExpiresIn=expires_in_sec,
            )
            return s3_uri.replace("https://", "s3://", 1)
        except Exception as e:
            logger.error(
                "failed to presign URI for get_object operation",
                uri=uri,
                exc_info=e,
                expires_in_sec=expires_in_sec,
            )
            raise

    async def upload(self, uri: str, value: bytes, logger: Any) -> None:
        """Stores the supplied binary value in a S3 object at the supplied URI.

        The URI must be S3 URI (starts with "s3://").
        Overwrites existing object. Raises Exception on error.
        """
        try:
            self._lazy_create_client()
            bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
            await asyncio.to_thread(
                self._s3_client.put_object, Bucket=bucket_name, Key=key, Body=value
            )
        except Exception as e:
            logger.error("failed to set S3 object", uri=uri, exc_info=e)
            raise

    async def create_multipart_upload(self, uri: str, logger: Any) -> str:
        """Creates a multipart upload for S3 object and returns the upload ID."""
        self._lazy_create_client()
        bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
        try:
            response = await asyncio.to_thread(
                self._s3_client.create_multipart_upload,
                Bucket=bucket_name,
                Key=key,
            )
            return response["UploadId"]
        except Exception as e:
            logger.error("failed to create multipart upload", uri=uri, exc_info=e)
            raise

    async def complete_multipart_upload(
        self, uri: str, upload_id: str, parts_etags: list[str], logger: Any
    ) -> None:
        """Completes a multipart upload for S3 object."""
        self._lazy_create_client()
        bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
        try:
            await asyncio.to_thread(
                self._s3_client.complete_multipart_upload,
                Bucket=bucket_name,
                Key=key,
                UploadId=upload_id,
                MultipartUpload={
                    "Parts": [
                        {"ETag": etag, "PartNumber": i + 1}
                        for i, etag in enumerate(parts_etags)
                    ]
                },
            )
        except Exception as e:
            logger.error("failed to complete multipart upload", uri=uri, exc_info=e)
            raise

    async def abort_multipart_upload(
        self, uri: str, upload_id: str, logger: Any
    ) -> None:
        """Aborts a multipart upload for S3 object."""
        self._lazy_create_client()
        bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
        try:
            await asyncio.to_thread(
                self._s3_client.abort_multipart_upload,
                Bucket=bucket_name,
                Key=key,
                UploadId=upload_id,
            )
        except Exception as e:
            logger.error("failed to abort multipart upload", uri=uri, exc_info=e)
            raise

    async def presign_upload_part_uri(
        self,
        uri: str,
        part_number: int,
        upload_id: str,
        expires_in_sec: int,
        logger: Any,
    ) -> str:
        """Returns a presigned URI for uploading a part in a multipart upload for S3 object."""
        self._lazy_create_client()
        bucket_name, key = _bucket_name_and_object_key_from_uri(uri)
        try:
            response = await asyncio.to_thread(
                self._s3_client.generate_presigned_url,
                ClientMethod="upload_part",
                Params={
                    "Bucket": bucket_name,
                    "Key": key,
                    "UploadId": upload_id,
                    "PartNumber": part_number,
                },
                ExpiresIn=expires_in_sec,
            )
            return response
        except Exception as e:
            logger.error(
                "failed to presign URI for upload_part operation",
                uri=uri,
                exc_info=e,
                part_number=part_number,
                upload_id=upload_id,
                expires_in_sec=expires_in_sec,
            )
            raise


def _bucket_name_and_object_key_from_uri(uri: str) -> tuple[str, str]:
    # Example S3 object URI:
    # s3://test-indexify-server-blob-store-eugene-20250411/225b83f4-2aed-40a7-adee-b7a681f817f2
    if not uri.startswith("s3://"):
        raise ValueError(f"S3 URI '{uri}' is missing 's3://' prefix")

    parts = uri[5:].split("/", 1)
    if len(parts) != 2:
        raise ValueError(f"Failed parsing bucket name from S3 URI '{uri}'")
    return parts[0], parts[1]  # bucket_name, key



from aiohttp import web

from indexify.proto.executor_api_pb2 import (
    DesiredExecutorState,
)

from ..state_reconciler import ExecutorStateReconciler
from .handler import Handler


class DesiredStateHandler(Handler):
    def __init__(self, state_reconciler: ExecutorStateReconciler):
        self._state_reconciler = state_reconciler

    async def handle(self, request: web.Request) -> web.Response:
        desired_state: DesiredExecutorState | None = (
            self._state_reconciler.get_desired_state()
        )
        if desired_state is None:
            return web.Response(
                status=200, text="No desired state received from Server yet"
            )
        else:
            return web.Response(text=str(desired_state))



import prometheus_client

from indexify.executor.monitoring.metrics import (
    latency_metric_for_customer_controlled_operation,
)

metric_allocations_completed: prometheus_client.Counter = prometheus_client.Counter(
    "allocations_completed",
    "Number of allocations that were completed",
    ["outcome_code", "failure_reason"],
)
METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_ALL = "all"
METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_SUCCESS = "success"
METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE = "failure"

METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALL = "all"
# Used when the allocation is successfull.
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_NONE = "none"
# Matches ALLOCATION_FAILURE_REASON_UNKNOWN
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_UNKNOWN = "unknown"
# Includes all function errors including timeouts to reduce cardinality.
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_ERROR = "function_error"
# Includes all internal errors to reduce cardinality.
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_INTERNAL_ERROR = "internal_error"
# Matches ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALLOCATION_CANCELLED = "alloc_cancelled"
# Matches ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED
METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED = (
    "function_executor_terminated"
)

# Valid combinations of the labels:
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_ALL,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALL,
)
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_SUCCESS,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_NONE,
)

metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_UNKNOWN,
)
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_ERROR,
)
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_INTERNAL_ERROR,
)
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALLOCATION_CANCELLED,
)
metric_allocations_completed.labels(
    outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED,
)

metric_allocation_completion_latency: prometheus_client.Histogram = (
    latency_metric_for_customer_controlled_operation(
        "allocation_completion",
        "allocation completion from the moment it got fetched until its output got uploaded to blob store",
    )
)



from tensorlake.function_executor.proto.function_executor_pb2 import FunctionInputs


class AllocationInput:
    """Represents the input for a task allocation in the function executor controller."""

    def __init__(
        self,
        function_inputs: FunctionInputs,
        function_outputs_blob_uri: str,
        function_outputs_blob_upload_id: str,
        request_error_blob_uri: str,
        request_error_blob_upload_id: str,
    ):
        # Actual input object sent to FE.
        self.function_inputs = function_inputs
        # Executor side function input related bookkeeping.
        self.function_outputs_blob_uri = function_outputs_blob_uri
        self.function_outputs_blob_upload_id = function_outputs_blob_upload_id
        self.request_error_blob_uri = request_error_blob_uri
        self.request_error_blob_upload_id = request_error_blob_upload_id



from enum import Enum
from typing import List

from indexify.executor.function_executor.function_executor import (
    FunctionExecutor,
)
from indexify.proto.executor_api_pb2 import FunctionExecutorTerminationReason

from .allocation_info import AllocationInfo


class EventType(Enum):
    FUNCTION_EXECUTOR_CREATED = 1
    FUNCTION_EXECUTOR_TERMINATED = 2
    SHUTDOWN_INITIATED = 3
    ALLOCATION_PREPARATION_FINISHED = 4
    SCHEDULE_ALLOCATION_EXECUTION = 5
    ALLOCATION_EXECUTION_FINISHED = 6
    ALLOCATION_FINALIZATION_FINISHED = 7


class BaseEvent:
    """
    Base class for events in the FunctionExecutorController.
    This class can be extended to create specific event types.
    """

    def __init__(self, event_type: EventType):
        self.event_type = event_type

    def __str__(self) -> str:
        return f"Event(type={self.event_type.name})"


class FunctionExecutorCreated(BaseEvent):
    """
    Event indicating that Function Executor got created or failed.

    The function_executor field is None if the function executor creation failed.
    In this case the fe_termination_reason field is set to the reason why.
    """

    def __init__(
        self,
        function_executor: FunctionExecutor | None,
        fe_termination_reason: FunctionExecutorTerminationReason | None,
    ):
        super().__init__(EventType.FUNCTION_EXECUTOR_CREATED)
        self.function_executor: FunctionExecutor | None = function_executor
        self.fe_termination_reason: FunctionExecutorTerminationReason | None = (
            fe_termination_reason
        )


class FunctionExecutorTerminated(BaseEvent):
    """
    Event indicating that Function Executor has been terminated (destroyed).
    """

    def __init__(
        self,
        is_success: bool,
        fe_termination_reason: FunctionExecutorTerminationReason,
        allocation_ids_caused_termination: List[str],
    ):
        super().__init__(EventType.FUNCTION_EXECUTOR_TERMINATED)
        self.is_success: bool = is_success
        self.fe_termination_reason: FunctionExecutorTerminationReason = (
            fe_termination_reason
        )
        self.allocation_ids_caused_termination: List[str] = (
            allocation_ids_caused_termination
        )

    def __str__(self) -> str:
        return (
            f"Event(type={self.event_type.name}, "
            f"is_success={self.is_success}, "
            f"fe_termination_reason={FunctionExecutorTerminationReason.Name(self.fe_termination_reason)}, "
            f"allocation_ids_caused_termination={self.allocation_ids_caused_termination})"
        )


class ShutdownInitiated(BaseEvent):
    """
    Event indicating that Function Executor shutdown has been initiated.
    """

    def __init__(self):
        super().__init__(EventType.SHUTDOWN_INITIATED)


class AllocationPreparationFinished(BaseEvent):
    """
    Event indicating that a allocation has been prepared for execution or failed to do that.
    """

    def __init__(
        self,
        alloc_info: AllocationInfo,
        is_success: bool,
    ):
        super().__init__(EventType.ALLOCATION_PREPARATION_FINISHED)
        self.alloc_info: AllocationInfo = alloc_info
        self.is_success: bool = is_success

    def __str__(self) -> str:
        return (
            f"Event(type={self.event_type.name}, "
            f"function_call_id={self.alloc_info.allocation.function_call_id}, "
            f"allocation_id={self.alloc_info.allocation.allocation_id}), "
            f"is_success={self.is_success}"
        )


class ScheduleAllocationExecution(BaseEvent):
    """
    Event indicating that a allocation has been scheduled.
    """

    def __init__(self):
        super().__init__(EventType.SCHEDULE_ALLOCATION_EXECUTION)


class AllocationExecutionFinished(BaseEvent):
    """
    Event indicating that a allocation execution has been finished on Function Executor.
    """

    def __init__(
        self,
        alloc_info: AllocationInfo,
        function_executor_termination_reason: (
            FunctionExecutorTerminationReason | None
        ),  # type: Optional[FunctionExecutorTerminationReason]
    ):
        super().__init__(EventType.ALLOCATION_EXECUTION_FINISHED)
        self.alloc_info: AllocationInfo = alloc_info
        # Not None if the FE needs to get destroyed after running the allocation.
        self.function_executor_termination_reason = function_executor_termination_reason

    def __str__(self) -> str:
        function_executor_termination_reason_str: str = (
            "None"
            if self.function_executor_termination_reason is None
            else FunctionExecutorTerminationReason.Name(
                self.function_executor_termination_reason
            )
        )
        return (
            f"Event(type={self.event_type.name}, "
            f"function_call_id={self.alloc_info.allocation.function_call_id}, "
            f"allocation_id={self.alloc_info.allocation.allocation_id}), "
            f"function_executor_termination_reason={function_executor_termination_reason_str}"
        )


class AllocationFinalizationFinished(BaseEvent):
    """
    Event indicating that an allocation finalization is finished.
    """

    def __init__(self, alloc_info: AllocationInfo, is_success: bool):
        super().__init__(EventType.ALLOCATION_FINALIZATION_FINISHED)
        self.alloc_info: AllocationInfo = alloc_info
        self.is_success: bool = is_success

    def __str__(self) -> str:
        return (
            f"Event(type={self.event_type.name}, "
            f"function_call_id={self.alloc_info.allocation.function_call_id}, "
            f"allocation_id={self.alloc_info.allocation.allocation_id}), "
            f"is_success={self.is_success}"
        )



from typing import Any

from indexify.proto.executor_api_pb2 import (
    Allocation,
    AllocationResult,
    FunctionExecutorDescription,
)


def function_executor_logger(
    function_executor_description: FunctionExecutorDescription, logger: Any
) -> Any:
    """Returns a logger bound with the FE's metadata.

    The function assumes that the FE might be invalid."""
    return logger.bind(
        fn_executor_id=(
            function_executor_description.id
            if function_executor_description.HasField("id")
            else None
        ),
        namespace=(
            function_executor_description.function.namespace
            if function_executor_description.HasField("function")
            and function_executor_description.function.HasField("namespace")
            else None
        ),
        app=(
            function_executor_description.function.application_name
            if function_executor_description.HasField("function")
            and function_executor_description.function.HasField("application_name")
            else None
        ),
        app_version=(
            function_executor_description.function.application_version
            if function_executor_description.HasField("function")
            and function_executor_description.function.HasField("application_version")
            else None
        ),
        fn=(
            function_executor_description.function.function_name
            if function_executor_description.HasField("function")
            and function_executor_description.function.HasField("function_name")
            else None
        ),
    )


def allocation_logger(alloc: Allocation, logger: Any) -> Any:
    """Returns a logger for the given Allocation.

    Doesn't assume that the supplied Allocation is valid.
    """
    return logger.bind(
        allocation_id=(
            alloc.allocation_id if alloc.HasField("allocation_id") else None
        ),
        fn_executor_id=(
            alloc.function_executor_id
            if alloc.HasField("function_executor_id")
            else None
        ),
        fn_call_id=(
            alloc.function_call_id if alloc.HasField("function_call_id") else None
        ),
        namespace=(
            alloc.function.namespace
            if alloc.HasField("function") and alloc.function.HasField("namespace")
            else None
        ),
        app=(
            alloc.function.application_name
            if alloc.HasField("function")
            and alloc.function.HasField("application_name")
            else None
        ),
        app_version=(
            alloc.function.application_version
            if alloc.HasField("function")
            and alloc.function.HasField("application_version")
            else None
        ),
        fn=(
            alloc.function.function_name
            if alloc.HasField("function") and alloc.function.HasField("function_name")
            else None
        ),
        request_id=(alloc.request_id if alloc.HasField("request_id") else None),
    )


def allocation_result_logger(alloc_result: AllocationResult, logger: Any) -> Any:
    """Returns a logger bound with the alloc result's metadata.

    The function assumes that the alloc result might be invalid."""
    return logger.bind(
        fn_call_id=(
            alloc_result.function_call_id
            if alloc_result.HasField("function_call_id")
            else None
        ),
        allocation_id=(
            alloc_result.allocation_id
            if alloc_result.HasField("allocation_id")
            else None
        ),
        namespace=(
            alloc_result.function.namespace
            if alloc_result.HasField("function")
            and alloc_result.function.HasField("namespace")
            else None
        ),
        app=(
            alloc_result.function.application_name
            if alloc_result.HasField("function")
            and alloc_result.function.HasField("application_name")
            else None
        ),
        app_version=(
            alloc_result.function.application_version
            if alloc_result.HasField("function")
            and alloc_result.function.HasField("application_version")
            else None
        ),
        fn=(
            alloc_result.function.function_name
            if alloc_result.HasField("function")
            and alloc_result.function.HasField("function_name")
            else None
        ),
        request_id=(
            alloc_result.request_id if alloc_result.HasField("request_id") else None
        ),
    )



from collections.abc import Iterable

from tensorlake.function_executor.proto.message_validator import MessageValidator

from indexify.proto.executor_api_pb2 import (
    Allocation,
    DataPayload,
    FunctionExecutorDescription,
    FunctionRef,
)


def validate_function_executor_description(
    function_executor_description: FunctionExecutorDescription,
) -> None:
    """Validates the supplied FE description.

    Raises ValueError if the description is not valid.
    """
    validator = MessageValidator(function_executor_description)
    validator.required_field("id")
    validator.required_field("function")
    _validate_function_ref(function_executor_description.function)
    # secret_names can be empty.
    validator.required_field("initialization_timeout_ms")
    validator.required_field("application")
    validator.required_field("resources")
    validator.required_field("output_payload_uri_prefix")
    validator.required_field("max_concurrency")
    validator.required_field("allocation_timeout_ms")

    _validate_data_payload(function_executor_description.application)

    validator = MessageValidator(function_executor_description.resources)
    validator.required_field("cpu_ms_per_sec")
    validator.required_field("memory_bytes")
    validator.required_field("disk_bytes")

    if function_executor_description.resources.HasField("gpu"):
        validator = MessageValidator(function_executor_description.resources.gpu)
        validator.required_field("count")
        validator.required_field("model")


def validate_allocation(alloc: Allocation) -> None:
    """Validates the supplied TaskAllocation.

    Raises ValueError if the TaskAllocation is not valid.
    """
    validator = MessageValidator(alloc)
    validator.required_field("function")
    _validate_function_ref(alloc.function)
    validator.required_field("allocation_id")
    validator.required_field("function_call_id")
    validator.required_field("request_id")
    _validate_data_payloads(alloc.args, field_name="TaskAllocation.args")
    validator.required_field("output_payload_uri_prefix")
    validator.required_field("request_error_payload_uri_prefix")
    validator.required_field("function_executor_id")
    validator.required_field("function_call_metadata")


def _validate_data_payloads(
    data_payloads: Iterable[DataPayload], field_name: str
) -> None:
    """Validates the supplied iterable of DataPayloads."""
    for data_payload in data_payloads:
        _validate_data_payload(data_payload)


def _validate_data_payload(data_payload: DataPayload) -> None:
    """Validates the supplied DataPayload.

    Raises ValueError if the DataPayload is not valid.
    """
    (
        MessageValidator(data_payload)
        .required_field("uri")
        .required_field("encoding")
        .required_field("encoding_version")
        # content_type is optional.
        # metadata_size is optional.
        .required_field("offset")
        .required_field("size")
        .required_field("sha256_hash")
        # source_function_call_id is not used
        # id is not used
    )


def _validate_function_ref(function_ref: FunctionRef) -> None:
    """Validates the supplied FunctionRef.

    Raises ValueError if the FunctionRef is not valid.
    """
    (
        MessageValidator(function_ref)
        .required_field("namespace")
        .required_field("application_name")
        .required_field("function_name")
        # application_version is not always set by Server.
    )



import prometheus_client

# This file contains all metrics used by Executor.

# Executor overview metrics.
metric_executor_info: prometheus_client.Info = prometheus_client.Info(
    "executor", "Executor information"
)
metric_executor_state: prometheus_client.Enum = prometheus_client.Enum(
    "executor_state",
    "Current Executor state",
    states=["starting", "running", "shutting_down"],
)

metric_executor_events_pushed: prometheus_client.Counter = prometheus_client.Counter(
    "executor_events_pushed", "Number of events pushed to collector"
)

metric_executor_event_push_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "executor_event_push_errors",
        "Number of errors while pushing events to collector",
    )
)



import prometheus_client

from indexify.executor.monitoring.metrics import latency_metric_for_fast_operation

# Application download metrics
metric_application_downloads: prometheus_client.Counter = prometheus_client.Counter(
    "application_downloads",
    "Number of application downloads, including downloads served from local cache",
)
metric_application_download_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "application_download_errors",
        "Number of application download errors, including downloads served from local cache",
    )
)
metric_applications_from_cache: prometheus_client.Counter = prometheus_client.Counter(
    "application_downloads_from_cache",
    "Number of application downloads served from local cache",
)
metric_application_download_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "application_download",
        "Application download, including downloads served from local cache",
    )
)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import warnings

import grpc

from indexify.proto import executor_api_pb2 as indexify_dot_proto_dot_executor__api__pb2

GRPC_GENERATED_VERSION = "1.75.1"
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in indexify/proto/executor_api_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
    )


class ExecutorAPIStub(object):
    """Internal API for scheduling and running allocations on Executors. Executors are acting as clients of this API.
    Server is responsible for scheduling allocations on Executors and Executors are responsible for running the allocations.

    Rename with caution. Existing clients won't find the service if the service name changes. A HTTP2 ingress proxy
    might use the service name in it HTTP2 path based routing rules. See how gRPC uses service names in its HTTP2 paths
    at https://github.com/grpc/grpc/blob/master/doc/PROTOCOL-HTTP2.md.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.report_executor_state = channel.unary_unary(
            "/executor_api_pb.ExecutorAPI/report_executor_state",
            request_serializer=indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateRequest.SerializeToString,
            response_deserializer=indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateResponse.FromString,
            _registered_method=True,
        )
        self.get_desired_executor_states = channel.unary_stream(
            "/executor_api_pb.ExecutorAPI/get_desired_executor_states",
            request_serializer=indexify_dot_proto_dot_executor__api__pb2.GetDesiredExecutorStatesRequest.SerializeToString,
            response_deserializer=indexify_dot_proto_dot_executor__api__pb2.DesiredExecutorState.FromString,
            _registered_method=True,
        )
        self.invoke_function = channel.unary_stream(
            "/executor_api_pb.ExecutorAPI/invoke_function",
            request_serializer=indexify_dot_proto_dot_executor__api__pb2.FunctionCallRequest.SerializeToString,
            response_deserializer=indexify_dot_proto_dot_executor__api__pb2.FunctionCallResponse.FromString,
            _registered_method=True,
        )


class ExecutorAPIServicer(object):
    """Internal API for scheduling and running allocations on Executors. Executors are acting as clients of this API.
    Server is responsible for scheduling allocations on Executors and Executors are responsible for running the allocations.

    Rename with caution. Existing clients won't find the service if the service name changes. A HTTP2 ingress proxy
    might use the service name in it HTTP2 path based routing rules. See how gRPC uses service names in its HTTP2 paths
    at https://github.com/grpc/grpc/blob/master/doc/PROTOCOL-HTTP2.md.
    """

    def report_executor_state(self, request, context):
        """Called by Executor every 5 seconds to report that it's still alive and provide its current state.

        Missing 3 reports will result in the Executor being deregistered by Server.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def get_desired_executor_states(self, request, context):
        """Called by Executor to open a stream of its desired states. When Server wants Executor to change something
        it puts a message on the stream with the new desired state of the Executor.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def invoke_function(self, request, context):
        """Called by the user code to invoke a blocking function call.
        returns a stream until the function call is completed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ExecutorAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "report_executor_state": grpc.unary_unary_rpc_method_handler(
            servicer.report_executor_state,
            request_deserializer=indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateRequest.FromString,
            response_serializer=indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateResponse.SerializeToString,
        ),
        "get_desired_executor_states": grpc.unary_stream_rpc_method_handler(
            servicer.get_desired_executor_states,
            request_deserializer=indexify_dot_proto_dot_executor__api__pb2.GetDesiredExecutorStatesRequest.FromString,
            response_serializer=indexify_dot_proto_dot_executor__api__pb2.DesiredExecutorState.SerializeToString,
        ),
        "invoke_function": grpc.unary_stream_rpc_method_handler(
            servicer.invoke_function,
            request_deserializer=indexify_dot_proto_dot_executor__api__pb2.FunctionCallRequest.FromString,
            response_serializer=indexify_dot_proto_dot_executor__api__pb2.FunctionCallResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "executor_api_pb.ExecutorAPI", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers(
        "executor_api_pb.ExecutorAPI", rpc_method_handlers
    )


# This class is part of an EXPERIMENTAL API.
class ExecutorAPI(object):
    """Internal API for scheduling and running allocations on Executors. Executors are acting as clients of this API.
    Server is responsible for scheduling allocations on Executors and Executors are responsible for running the allocations.

    Rename with caution. Existing clients won't find the service if the service name changes. A HTTP2 ingress proxy
    might use the service name in it HTTP2 path based routing rules. See how gRPC uses service names in its HTTP2 paths
    at https://github.com/grpc/grpc/blob/master/doc/PROTOCOL-HTTP2.md.
    """

    @staticmethod
    def report_executor_state(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/executor_api_pb.ExecutorAPI/report_executor_state",
            indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateRequest.SerializeToString,
            indexify_dot_proto_dot_executor__api__pb2.ReportExecutorStateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def get_desired_executor_states(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/executor_api_pb.ExecutorAPI/get_desired_executor_states",
            indexify_dot_proto_dot_executor__api__pb2.GetDesiredExecutorStatesRequest.SerializeToString,
            indexify_dot_proto_dot_executor__api__pb2.DesiredExecutorState.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def invoke_function(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/executor_api_pb.ExecutorAPI/invoke_function",
            indexify_dot_proto_dot_executor__api__pb2.FunctionCallRequest.SerializeToString,
            indexify_dot_proto_dot_executor__api__pb2.FunctionCallResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )



# We send function inputs and outputs over gRPC.
# -1 means unlimited. We don't want to limit the size of data customers are using.
# The effective max message size in this case is about 1.9 GB, see the max payload test.
# This is due to internal hard gRPC limits. When we want to increase the message sizes
# we'll have to implement chunking for large messages.
_MAX_GRPC_MESSAGE_LENGTH = -1

# Optimize the channels for low latency connection establishement as we are running on the same host.
_RECONNECT_BACKOFF_MS = 100

GRPC_CHANNEL_OPTIONS = [
    ("grpc.max_receive_message_length", _MAX_GRPC_MESSAGE_LENGTH),
    ("grpc.max_send_message_length", _MAX_GRPC_MESSAGE_LENGTH),
    ("grpc.min_reconnect_backoff_ms", _RECONNECT_BACKOFF_MS),
    ("grpc.max_reconnect_backoff_ms", _RECONNECT_BACKOFF_MS),
    ("grpc.initial_reconnect_backoff_ms", _RECONNECT_BACKOFF_MS),
]

# If a health check takes more than this duration then the server is considered unhealthy.
HEALTH_CHECK_TIMEOUT_SEC = 5



import asyncio
from pathlib import Path
from typing import Any, Tuple

from tensorlake.function_executor.proto.function_executor_pb2 import (
    FunctionRef,
    InitializationFailureReason,
    InitializationOutcomeCode,
    InitializeRequest,
    InitializeResponse,
    SerializedObject,
)
from tensorlake.function_executor.proto.message_validator import MessageValidator

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.executor.function_executor.function_executor import (
    FunctionExecutor,
    FunctionExecutorInitializationResult,
)
from indexify.executor.function_executor.server.function_executor_server_factory import (
    FunctionExecutorServerConfiguration,
    FunctionExecutorServerFactory,
)
from indexify.proto.executor_api_pb2 import (
    FunctionExecutorDescription,
    FunctionExecutorTerminationReason,
)

from .aio_utils import shielded_await
from .downloads import download_application_code
from .events import FunctionExecutorCreated


async def create_function_executor(
    function_executor_description: FunctionExecutorDescription,
    function_executor_server_factory: FunctionExecutorServerFactory,
    blob_store: BLOBStore,
    executor_id: str,
    base_url: str,
    config_path: str,
    cache_path: Path,
    logger: Any,
) -> FunctionExecutorCreated:
    """Creates a function executor.

    Doesn't raise any exceptions.
    """
    logger = logger.bind(module=__name__)
    try:
        function_executor, result = await _create_function_executor(
            function_executor_description=function_executor_description,
            function_executor_server_factory=function_executor_server_factory,
            blob_store=blob_store,
            executor_id=executor_id,
            base_url=base_url,
            config_path=config_path,
            cache_path=cache_path,
            logger=logger,
        )
    except asyncio.CancelledError:
        # Cancelled FE startup means that Server removed this FE from desired state. We don't have FE termination reason for the case
        # when Server removed FE from desired state because we can't rely on its delivery because FE removed from desired state can get
        # removed from reported state at any moment. Thus we can use any termination reason here.
        return FunctionExecutorCreated(
            function_executor=None,
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED,
        )
    except BaseException as e:
        logger.error(
            "failed to create function executor",
            exc_info=e,
        )
        return FunctionExecutorCreated(
            function_executor=None,
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR,
        )

    function_executor: FunctionExecutor
    result: FunctionExecutorInitializationResult
    # No await here so this call can't be cancelled.
    fe_created_event: FunctionExecutorCreated = _to_fe_created_event(
        function_executor=function_executor,
        result=result,
        logger=logger,
    )
    if fe_created_event.function_executor is None:
        # _to_fe_created_event doesn't like the FE, destroy it.
        fe_destroy_task: asyncio.Task = asyncio.create_task(
            function_executor.destroy(),
            name=f"destroy function executor {function_executor_description.id}",
        )
        try:
            await shielded_await(fe_destroy_task, logger)
        except asyncio.CancelledError:
            # destroy() finished due to the shield, return fe_created_event.
            pass

    return fe_created_event


def _to_fe_created_event(
    function_executor: FunctionExecutor,
    result: FunctionExecutorInitializationResult,
    logger: Any,
) -> FunctionExecutorCreated:
    """Converts FunctionExecutorInitializationResult to FunctionExecutorCreated event.

    Doesn't raise any exceptions.
    """
    if result.is_timeout:
        return FunctionExecutorCreated(
            function_executor=None,
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT,
        )

    if result.response is None:
        # This is a grey failure where we don't know the exact cause.
        # Treat it as a customer function error to prevent service abuse by intentionally
        # triggering function executor creations failures that don't get billed.
        logger.error("function executor startup failed with no response")
        return FunctionExecutorCreated(
            function_executor=None,
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR,
        )

    initialize_response: InitializeResponse = result.response
    try:
        _validate_initialize_response(initialize_response)
    except ValueError as e:
        # Grey failure mode. Treat as customer function error to prevent service abuse but log for future investigations.
        logger.error(
            "function executor initialization failed with invalid response", exc_info=e
        )
        return FunctionExecutorCreated(
            function_executor=None,
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR,
        )

    if (
        initialize_response.outcome_code
        == InitializationOutcomeCode.INITIALIZATION_OUTCOME_CODE_SUCCESS
    ):
        return FunctionExecutorCreated(
            function_executor=function_executor,
            fe_termination_reason=None,
        )
    else:
        if (
            initialize_response.failure_reason
            == InitializationFailureReason.INITIALIZATION_FAILURE_REASON_FUNCTION_ERROR
        ):
            return FunctionExecutorCreated(
                function_executor=None,
                fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR,
            )
        else:
            # Treat all other failure reasons as grey failures. Report them as function errors to prevent service abuse.
            # Log them for awareness and future investigations.
            logger.error(
                "function executor initialization failed",
                failure_reason=InitializationFailureReason.Name(
                    initialize_response.failure_reason
                ),
            )
            return FunctionExecutorCreated(
                function_executor=None,
                fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR,
            )


async def _create_function_executor(
    function_executor_description: FunctionExecutorDescription,
    function_executor_server_factory: FunctionExecutorServerFactory,
    blob_store: BLOBStore,
    executor_id: str,
    base_url: str,
    config_path: str,
    cache_path: Path,
    logger: Any,
) -> Tuple[FunctionExecutor, FunctionExecutorInitializationResult]:
    """Creates a function executor.

    Raises Exception on internal Executor error.
    """
    application_code: SerializedObject = await download_application_code(
        function_executor_description=function_executor_description,
        cache_path=cache_path,
        blob_store=blob_store,
        logger=logger,
    )

    gpu_count: int = 0
    if function_executor_description.resources.HasField("gpu"):
        gpu_count = function_executor_description.resources.gpu.count

    config: FunctionExecutorServerConfiguration = FunctionExecutorServerConfiguration(
        executor_id=executor_id,
        function_executor_id=function_executor_description.id,
        namespace=function_executor_description.function.namespace,
        application_name=function_executor_description.function.application_name,
        application_version=function_executor_description.function.application_version,
        function_name=function_executor_description.function.function_name,
        secret_names=list(function_executor_description.secret_names),
        cpu_ms_per_sec=function_executor_description.resources.cpu_ms_per_sec,
        memory_bytes=function_executor_description.resources.memory_bytes,
        disk_bytes=function_executor_description.resources.disk_bytes,
        gpu_count=gpu_count,
    )

    initialize_request: InitializeRequest = InitializeRequest(
        function=FunctionRef(
            namespace=function_executor_description.function.namespace,
            application_name=function_executor_description.function.application_name,
            application_version=function_executor_description.function.application_version,
            function_name=function_executor_description.function.function_name,
        ),
        application_code=application_code,
    )
    customer_code_timeout_sec: float = (
        function_executor_description.initialization_timeout_ms / 1000.0
    )

    function_executor: FunctionExecutor = FunctionExecutor(
        server_factory=function_executor_server_factory, logger=logger
    )

    try:
        result: FunctionExecutorInitializationResult = (
            await function_executor.initialize(
                config=config,
                initialize_request=initialize_request,
                base_url=base_url,
                config_path=config_path,
                customer_code_timeout_sec=customer_code_timeout_sec,
            )
        )
        return (function_executor, result)
    except BaseException:
        fe_destroy_task: asyncio.Task = asyncio.create_task(
            function_executor.destroy(),
            name=f"destroy function executor {function_executor_description.id}",
        )
        # This await is a cancellation point, need to shield to ensure we destroyed the FE.
        await shielded_await(
            fe_destroy_task,
            logger,
        )
        raise


def _validate_initialize_response(
    response: InitializeResponse,
) -> None:
    """Validates the initialization response.

    Raises ValueError if the response is not valid.
    """
    validator: MessageValidator = MessageValidator(response)
    (validator.required_field("outcome_code"))
    if (
        response.outcome_code
        == InitializationOutcomeCode.INITIALIZATION_OUTCOME_CODE_FAILURE
    ):
        validator.required_field("failure_reason")

    if response.outcome_code not in [
        InitializationOutcomeCode.INITIALIZATION_OUTCOME_CODE_SUCCESS,
        InitializationOutcomeCode.INITIALIZATION_OUTCOME_CODE_FAILURE,
    ]:
        raise ValueError(f"Invalid outcome code: {response.outcome_code}")

    if response.failure_reason not in [
        InitializationFailureReason.INITIALIZATION_FAILURE_REASON_UNKNOWN,
        InitializationFailureReason.INITIALIZATION_FAILURE_REASON_FUNCTION_ERROR,
        InitializationFailureReason.INITIALIZATION_FAILURE_REASON_INTERNAL_ERROR,
    ]:
        raise ValueError(f"Invalid failure reason: {response.failure_reason}")



from typing import Optional

from .health_checker import HealthChecker, HealthCheckResult
from .metrics.health_checker import metric_healthy

_HEALTH_CHECKER_NAME = "GenericHealthChecker"


class GenericHealthChecker(HealthChecker):
    """A generic health checker that doesn't depend on machine type and other features of the environment.

    The health checker uses software signals available in all environments like Function Executor failure rates.
    """

    def __init__(self):
        self._server_connection_unhealthy_status_message: Optional[str] = None
        metric_healthy.set(1)

    def server_connection_state_changed(self, is_healthy: bool, status_message: str):
        """Handle changes in server connection state."""
        if is_healthy:
            self._server_connection_unhealthy_status_message = None
            metric_healthy.set(1)
        else:
            self._server_connection_unhealthy_status_message = status_message
            metric_healthy.set(0)

    async def check(self) -> HealthCheckResult:
        if self._server_connection_unhealthy_status_message is not None:
            return HealthCheckResult(
                is_success=False,
                status_message=self._server_connection_unhealthy_status_message,
                checker_name=_HEALTH_CHECKER_NAME,
            )

        return HealthCheckResult(
            is_success=True,
            status_message="Successful",
            checker_name=_HEALTH_CHECKER_NAME,
        )



from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class FunctionURI:
    namespace: str
    application: str
    compute_fn: str
    version: Optional[str] = None


def function_allowlist_to_indexed_dict(
    function_allowlist: List[FunctionURI],
) -> Dict[str, str]:
    """Returns a dictionary with each function URI in the allowlist as a key-value pair.

    The keys are prefixed indexes in function allowlist, and the values are the function URIs
    """
    indexed_dict = {}
    counter = 0
    for function_uri in function_allowlist:
        function_uri: FunctionURI
        indexed_dict[f"function_allowlist_{counter}"] = ":".join(
            [
                function_uri.namespace,
                function_uri.application,
                function_uri.compute_fn,
                str(function_uri.version),
            ]
        )
        counter += 1
    return indexed_dict


def parse_function_uris(function_uri_strs: List[str]) -> List[FunctionURI]:
    """Parses a list of function URIs from strings to FunctionURI objects."""
    uris: List[FunctionURI] = []
    for uri_str in function_uri_strs:
        tokens = uri_str.split(":")
        if len(tokens) < 3 or len(tokens) > 4:
            raise ValueError(
                "Function should be specified as <namespace>:<application>:<function>:<version> or"
                "<namespace>:<application>:<function>"
            )
        version: Optional[str] = None
        if len(tokens) == 4:
            version = tokens[3]

        uris.append(
            FunctionURI(
                namespace=tokens[0],
                application=tokens[1],
                compute_fn=tokens[2],
                version=version,
            )
        )

    return uris



from aiohttp import web

from .handler import Handler


class StartupProbeHandler(Handler):
    def __init__(self):
        self._ready = False

    def set_ready(self):
        self._ready = True

    async def handle(self, request: web.Request) -> web.Response:
        if self._ready:
            return web.json_response({"status": "ok"})
        else:
            return web.json_response({"status": "nok"}, status=503)



import prometheus_client

from ...monitoring.metrics import latency_metric_for_fast_operation

# This file contains all metrics used by RequestStateClient.

# General metrics.
metric_request_read_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_request_state_client_request_read_errors",
    "Number of failed request reads in Function Executor Request State client resulting in its early termination",
)

# Get request state key-value Server API metrics.
metric_server_get_state_requests: prometheus_client.Counter = prometheus_client.Counter(
    "server_get_request_state_requests",
    "Number of get request state requests sent to the Server on behalf of Function Executor",
)
metric_server_get_state_request_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "server_get_request_state_request_errors",
        "Server get request state request errors",
    )
)
metric_server_get_state_request_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "server_get_request_state_request", "Server get request state request"
    )
)

# Set request state key-value Server API metrics.
metric_server_set_state_requests: prometheus_client.Counter = prometheus_client.Counter(
    "server_set_request_state_requests",
    "Number of set request state requests sent to the Server on behalf of Function Executor",
)
metric_server_set_state_request_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "server_set_request_state_request_errors",
        "Server set request state request errors",
    )
)
metric_server_set_state_request_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "server_set_request_state_request", "Server set request state request"
    )
)



from aiohttp import web

from indexify.proto.executor_api_pb2 import (
    ReportExecutorStateRequest,
)

from ..state_reporter import ExecutorStateReporter
from .handler import Handler


class ReportedStateHandler(Handler):
    def __init__(self, state_reporter: ExecutorStateReporter):
        self._state_reporter = state_reporter

    async def handle(self, request: web.Request) -> web.Response:
        request: ReportExecutorStateRequest | None = (
            self._state_reporter.last_state_report_request()
        )
        if request is None:
            return web.Response(status=200, text="No state reported so far")
        else:
            return web.Response(text=str(request))



import prometheus_client

from ...monitoring.metrics import latency_metric_for_fast_operation

metric_get_blob_requests: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_get_requests",
    "Number of get BLOB requests in BLOB store",
)
metric_get_blob_errors: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_get_request_errors",
    "Number of get BLOB request errors in BLOB store",
)
metric_get_blob_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_get",
        "BLOB store get BLOB request",
    )
)

metric_presign_uri_requests: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_presign_uri_requests",
    "Number of presign URI requests in BLOB store",
)
metric_presign_uri_errors: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_presign_uri_request_errors",
    "Number of presign URI request errors in BLOB store",
)
metric_presign_uri_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_presign_uri",
        "BLOB store presign URI request",
    )
)

metric_upload_blob_requests: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_upload_requests",
    "Number of upload BLOB requests in BLOB store",
)
metric_upload_blob_errors: prometheus_client.Counter = prometheus_client.Counter(
    "blob_store_upload_request_errors",
    "Number of upload BLOB request errors in BLOB store",
)
metric_upload_blob_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_upload",
        "BLOB store upload BLOB request",
    )
)

metric_create_multipart_upload_requests: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_create_multipart_upload_requests",
        "Number of create multipart upload requests in BLOB store",
    )
)
metric_create_multipart_upload_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_create_multipart_upload_request_errors",
        "Number of create multipart upload request errors in BLOB store",
    )
)
metric_create_multipart_upload_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_create_multipart_upload_request",
        "create multipart upload request in BLOB store",
    )
)

metric_complete_multipart_upload_requests: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_complete_multipart_upload_requests",
        "Number of complete multipart upload requests in BLOB store",
    )
)
metric_complete_multipart_upload_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_complete_multipart_upload_request_errors",
        "Number of complete multipart upload request errors in BLOB store",
    )
)
metric_complete_multipart_upload_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_complete_multipart_upload_request",
        "complete multipart upload request in BLOB store",
    )
)

metric_abort_multipart_upload_requests: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_abort_multipart_upload_requests",
        "Number of abort multipart upload requests in BLOB store",
    )
)
metric_abort_multipart_upload_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "blob_store_abort_multipart_upload_request_errors",
        "Number of abort multipart upload request errors in BLOB store",
    )
)
metric_abort_multipart_upload_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "blob_store_abort_multipart_upload_request",
        "abort multipart upload request in BLOB store",
    )
)



import asyncio

from aiohttp import web
from prometheus_client import CONTENT_TYPE_LATEST, generate_latest

from .handler import Handler


class PrometheusMetricsHandler(Handler):
    async def handle(self, request: web.Request) -> web.Response:
        # Run the synchronous metrics generation code in ThreadPool thread
        # to not block the main asyncio loop.
        return await asyncio.to_thread(self._handle_sync)

    def _handle_sync(self) -> web.Response:
        return web.Response(
            body=generate_latest(), headers={"Content-Type": CONTENT_TYPE_LATEST}
        )



from aiohttp import web

from .handler import Handler
from .health_checker.health_checker import HealthChecker, HealthCheckResult


class HealthCheckHandler(Handler):
    def __init__(self, health_checker: HealthChecker):
        self._health_checker = health_checker

    async def handle(self, request: web.Request) -> web.Response:
        result: HealthCheckResult = await self._health_checker.check()
        return web.json_response(
            {
                "status": "ok" if result.is_success else "nok",
                "message": result.status_message,
                "checker": result.checker_name,
            },
            status=200 if result.is_success else 503,
        )



import time
from typing import Any

from indexify.proto.executor_api_pb2 import (
    AllocationFailureReason,
    AllocationOutcomeCode,
)

from .allocation_info import AllocationInfo
from .metrics.completed_allocation_metrics import (
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALL,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALLOCATION_CANCELLED,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_ERROR,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_INTERNAL_ERROR,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_NONE,
    METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_UNKNOWN,
    METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_ALL,
    METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
    METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_SUCCESS,
    metric_allocation_completion_latency,
    metric_allocations_completed,
)


def emit_completed_allocation_metrics(alloc_info: AllocationInfo, logger: Any) -> None:
    """Emits Prometheus metrics for a completed allocation.

    Doesn't raise any exceptions.
    """
    logger = logger.bind(module=__name__)
    metric_allocation_completion_latency.observe(
        time.monotonic() - alloc_info.start_time
    )

    alloc_outcome_code: AllocationOutcomeCode = alloc_info.output.outcome_code
    alloc_failure_reason: AllocationFailureReason = alloc_info.output.failure_reason
    metric_allocations_completed.labels(
        outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_ALL,
        failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALL,
    ).inc()
    if alloc_outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS:
        metric_allocations_completed.labels(
            outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_SUCCESS,
            failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_NONE,
        ).inc()
    elif alloc_outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE:
        if (
            alloc_failure_reason
            == AllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR
        ):
            metric_allocations_completed.labels(
                outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
                failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_INTERNAL_ERROR,
            ).inc()
        elif (
            alloc_failure_reason
            == AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED
        ):
            metric_allocations_completed.labels(
                outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
                failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED,
            ).inc()
        elif (
            alloc_failure_reason
            == AllocationFailureReason.ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED
        ):
            metric_allocations_completed.labels(
                outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
                failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_ALLOCATION_CANCELLED,
            ).inc()
        elif alloc_failure_reason in [
            AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_ERROR,
            AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT,
            AllocationFailureReason.ALLOCATION_FAILURE_REASON_REQUEST_ERROR,
        ]:
            metric_allocations_completed.labels(
                outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
                failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_FUNCTION_ERROR,
            ).inc()
        else:
            metric_allocations_completed.labels(
                outcome_code=METRIC_ALLOCATIONS_COMPLETED_OUTCOME_CODE_FAILURE,
                failure_reason=METRIC_ALLOCATIONS_COMPLETED_FAILURE_REASON_UNKNOWN,
            ).inc()
            logger.warning(
                "unexpected allocation failure reason",
                failure_reason=AllocationFailureReason.Name(alloc_failure_reason),
            )



import asyncio
import hashlib
import platform
import sys
from socket import gethostname
from typing import Any, Dict, List

from indexify.proto.executor_api_pb2 import (
    AllocationFailureReason,
    AllocationOutcomeCode,
    AllocationResult,
    AllowedFunction,
    ExecutorState,
    ExecutorStatus,
    ExecutorUpdate,
    FunctionExecutorState,
    GPUModel,
    GPUResources,
)
from indexify.proto.executor_api_pb2 import HostResources as HostResourcesProto
from indexify.proto.executor_api_pb2 import (
    ReportExecutorStateRequest,
)
from indexify.proto.executor_api_pb2_grpc import ExecutorAPIStub

from .channel_manager import ChannelManager
from .function_allowlist import FunctionURI
from .function_executor_controller.loggers import allocation_result_logger
from .host_resources.host_resources import HostResources, HostResourcesProvider
from .host_resources.nvidia_gpu import NVIDIA_GPU_MODEL
from .metrics.state_reporter import (
    metric_state_report_rpc_errors,
    metric_state_report_rpc_latency,
    metric_state_report_rpcs,
)
from .monitoring.health_checker.health_checker import HealthChecker

_REPORTING_INTERVAL_SEC = 5
_REPORTING_BACKOFF_SEC = 5
_REPORT_RPC_TIMEOUT_SEC = 5


class ExecutorStateReporter:
    def __init__(
        self,
        executor_id: str,
        version: str,
        labels: Dict[str, str],
        function_allowlist: List[FunctionURI],
        channel_manager: ChannelManager,
        host_resources_provider: HostResourcesProvider,
        health_checker: HealthChecker,
        logger: Any,
    ):
        self._executor_id: str = executor_id
        self._version: str = version
        self._labels: Dict[str, str] = labels.copy()
        self._labels.update(_executor_labels())
        self._hostname: str = gethostname()
        self._channel_manager = channel_manager
        self._health_checker: HealthChecker = health_checker
        self._logger: Any = logger.bind(module=__name__)
        self._allowed_functions: List[AllowedFunction] = _to_allowed_function_protos(
            function_allowlist
        )
        # We need to fetch total resources only once, because they are not changing.
        self._total_host_resources: HostResources = (
            host_resources_provider.total_host_resources(self._logger)
        )
        self._total_function_executor_resources: HostResources = (
            host_resources_provider.total_function_executor_resources(self._logger)
        )
        self._logger.info(
            "detected host resources",
            total_host_resources=self._total_host_resources,
            total_function_executor_resources=self._total_function_executor_resources,
        )
        self._state_report_worker: asyncio.Task | None = None
        self._periodic_state_report_scheduler: asyncio.Task | None = None

        # Mutable fields
        self._state_report_scheduled_event: asyncio.Event = asyncio.Event()
        self._state_reported_event: asyncio.Event = asyncio.Event()
        self._executor_status: ExecutorStatus = ExecutorStatus.EXECUTOR_STATUS_UNKNOWN
        self._last_server_clock: int = (
            0  # Server expects initial value to be 0 until it is set by Server.
        )
        self._pending_allocation_results: List[AllocationResult] = []
        self._function_executor_states: Dict[str, FunctionExecutorState] = {}
        self._last_state_report_request: ReportExecutorStateRequest | None = None

    def last_state_report_request(self) -> ReportExecutorStateRequest | None:
        return self._last_state_report_request

    def update_executor_status(self, value: ExecutorStatus) -> None:
        self._executor_status = value

    def update_last_server_clock(self, value: int) -> None:
        self._last_server_clock = value

    def update_function_executor_state(
        self,
        state: FunctionExecutorState,
    ) -> None:
        self._function_executor_states[state.description.id] = state

    def remove_function_executor_state(self, function_executor_id: str) -> None:
        if function_executor_id not in self._function_executor_states:
            self._logger.warning(
                "attempted to remove non-existing function executor state",
                function_executor_id=function_executor_id,
            )
            return

        self._function_executor_states.pop(function_executor_id)

    def add_completed_allocation_result(self, alloc_result: AllocationResult) -> None:
        self._pending_allocation_results.append(alloc_result)

    def schedule_state_report(self) -> None:
        """Schedules a state report to be sent to the server asap.

        This method is called when the executor state changes and it needs to get reported.
        The call doesn't block and returns immediately.
        """
        self._state_report_scheduled_event.set()

    async def report_state_and_wait_for_completion(self) -> None:
        """Schedules a state report to be sent to the server asap and waits for the completion of the reporting."""
        self._state_reported_event.clear()
        self.schedule_state_report()
        await self._state_reported_event.wait()

    def run(self) -> None:
        """Runs the state reporter.

        This method is called when the executor starts and it needs to start reporting its state
        periodically.
        """
        self._state_report_worker = asyncio.create_task(
            self._state_report_worker_loop(), name="state_reporter_worker"
        )
        self._periodic_state_report_scheduler = asyncio.create_task(
            self._periodic_state_report_scheduler_loop(),
            name="state_reporter_periodic_scheduler",
        )

    async def shutdown(self) -> None:
        """Tries to do one last state report and shuts down the state reporter.

        Doesn't raise any exceptions."""
        if self._state_report_worker is not None:
            self._state_report_worker.cancel()
            try:
                await self._state_report_worker
            except asyncio.CancelledError:
                pass  # Expected exception
            self._state_report_worker = None

        if self._periodic_state_report_scheduler is not None:
            self._periodic_state_report_scheduler.cancel()
            try:
                await self._periodic_state_report_scheduler
            except asyncio.CancelledError:
                pass
            self._periodic_state_report_scheduler = None

        # Don't retry state report if it failed during shutdown.
        # We only do best effort last state report and Server might not be available.
        try:
            async with self._channel_manager.create_standalone_channel() as channel:
                await ExecutorAPIStub(channel).report_executor_state(
                    ReportExecutorStateRequest(
                        executor_state=self._current_executor_state(),
                        executor_update=self._remove_pending_update(),
                    ),
                    timeout=_REPORT_RPC_TIMEOUT_SEC,
                )
        except Exception as e:
            self._logger.error(
                "failed to report state during shutdown",
                exc_info=e,
            )

    async def _periodic_state_report_scheduler_loop(self) -> None:
        while True:
            self._state_report_scheduled_event.set()
            await asyncio.sleep(_REPORTING_INTERVAL_SEC)

    async def _state_report_worker_loop(self) -> None:
        """Runs the state reporter.

        Never raises any exceptions.
        """
        while True:
            stub = ExecutorAPIStub(await self._channel_manager.get_shared_channel())
            while True:
                await self._state_report_scheduled_event.wait()
                # Clear the event immidiately to report again asap if needed. This reduces latency in the system.
                self._state_report_scheduled_event.clear()
                try:
                    state: ExecutorState = self._current_executor_state()
                    update: ExecutorUpdate = self._remove_pending_update()
                    request: ReportExecutorStateRequest = ReportExecutorStateRequest(
                        executor_state=state, executor_update=update
                    )
                    _log_reported_executor_update(update, self._logger)
                    self._last_state_report_request = request

                    with (
                        metric_state_report_rpc_errors.count_exceptions(),
                        metric_state_report_rpc_latency.time(),
                    ):
                        metric_state_report_rpcs.inc()
                        await stub.report_executor_state(
                            request, timeout=_REPORT_RPC_TIMEOUT_SEC
                        )
                    self._state_reported_event.set()
                    self._health_checker.server_connection_state_changed(
                        is_healthy=True, status_message="grpc server channel is healthy"
                    )
                except Exception as e:
                    self._add_to_pending_update(update)
                    self._logger.error(
                        f"failed to report state to the server, backing-off for {_REPORTING_BACKOFF_SEC} sec.",
                        exc_info=e,
                    )
                    # The periodic state reports serve as channel health monitoring requests
                    # (same as TCP keep-alive). Channel Manager returns the same healthy channel
                    # for all RPCs that we do from Executor to Server. So all the RPCs benefit
                    # from this channel health monitoring.
                    self._health_checker.server_connection_state_changed(
                        is_healthy=False,
                        status_message="grpc server channel is unhealthy",
                    )
                    await self._channel_manager.fail_shared_channel()
                    await asyncio.sleep(_REPORTING_BACKOFF_SEC)
                    break  # exit the inner loop to use the recreated channel

    def _current_executor_state(self) -> ExecutorState:
        """Returns the current executor state."""
        state = ExecutorState(
            executor_id=self._executor_id,
            hostname=self._hostname,
            version=self._version,
            status=self._executor_status,
            total_function_executor_resources=_to_host_resources_proto(
                self._total_function_executor_resources
            ),
            total_resources=_to_host_resources_proto(self._total_host_resources),
            allowed_functions=self._allowed_functions,
            function_executor_states=list(self._function_executor_states.values()),
            labels=self._labels,
        )
        state.state_hash = _state_hash(state)
        # Set fields not included in the state hash.
        state.server_clock = self._last_server_clock
        return state

    def _remove_pending_update(self) -> ExecutorUpdate:
        """Removes all pending executor updates and returns them."""
        # No races here cause we don't await.
        alloc_results: List[AllocationResult] = self._pending_allocation_results
        self._pending_allocation_results = []

        return ExecutorUpdate(
            executor_id=self._executor_id,
            allocation_results=alloc_results,
        )

    def _add_to_pending_update(self, update: ExecutorUpdate) -> None:
        for alloc_result in update.allocation_results:
            self.add_completed_allocation_result(alloc_result)


def _log_reported_executor_update(update: ExecutorUpdate, logger: Any) -> None:
    """Logs the reported executor update.

    Doesn't raise any exceptions."""
    try:
        for alloc_result in update.allocation_results:
            allocation_result_logger(alloc_result, logger).info(
                "reporting allocation outcome",
                outcome_code=AllocationOutcomeCode.Name(alloc_result.outcome_code),
                failure_reason=(
                    AllocationFailureReason.Name(alloc_result.failure_reason)
                    if alloc_result.HasField("failure_reason")
                    else "None"
                ),
            )
    except Exception as e:
        logger.error(
            "failed to log reported executor update",
            exc_info=e,
        )


def _to_allowed_function_protos(
    function_allowlist: List[FunctionURI],
) -> List[AllowedFunction]:
    allowed_functions: List[AllowedFunction] = []
    for function_uri in function_allowlist:
        function_uri: FunctionURI
        allowed_function = AllowedFunction(
            namespace=function_uri.namespace,
            application_name=function_uri.application,
            function_name=function_uri.compute_fn,
        )
        if function_uri.version is not None:
            allowed_function.application_version = function_uri.version
        allowed_functions.append(allowed_function)

    return allowed_functions


def _state_hash(state: ExecutorState) -> str:
    serialized_state: bytes = state.SerializeToString(deterministic=True)
    hasher = hashlib.sha256(usedforsecurity=False)
    hasher.update(serialized_state)
    return hasher.hexdigest()


def _to_host_resources_proto(host_resources: HostResources) -> HostResourcesProto:
    proto = HostResourcesProto(
        cpu_count=host_resources.cpu_count,
        memory_bytes=host_resources.memory_mb * 1024 * 1024,
        disk_bytes=host_resources.disk_mb * 1024 * 1024,
    )
    if len(host_resources.gpus) > 0:
        proto.gpu.CopyFrom(
            GPUResources(
                count=len(host_resources.gpus),
                model=_to_gpu_model_proto(
                    host_resources.gpus[0].model
                ),  # All GPUs have the same model
            )
        )
    return proto


def _to_gpu_model_proto(nvidia_gpu_model: NVIDIA_GPU_MODEL) -> GPUModel:
    if nvidia_gpu_model == NVIDIA_GPU_MODEL.A100_40GB:
        return GPUModel.GPU_MODEL_NVIDIA_A100_40GB
    elif nvidia_gpu_model == NVIDIA_GPU_MODEL.A100_80GB:
        return GPUModel.GPU_MODEL_NVIDIA_A100_80GB
    elif nvidia_gpu_model == NVIDIA_GPU_MODEL.H100_80GB:
        return GPUModel.GPU_MODEL_NVIDIA_H100_80GB
    elif nvidia_gpu_model == NVIDIA_GPU_MODEL.TESLA_T4:
        return GPUModel.GPU_MODEL_NVIDIA_TESLA_T4
    elif nvidia_gpu_model == NVIDIA_GPU_MODEL.A6000:
        return GPUModel.GPU_MODEL_NVIDIA_A6000
    elif nvidia_gpu_model == NVIDIA_GPU_MODEL.A10:
        return GPUModel.GPU_MODEL_NVIDIA_A10
    else:
        return GPUModel.GPU_MODEL_UNKNOWN


def _executor_labels() -> Dict[str, str]:
    """Returns standard executor labels always added to user supplied labels."""
    return {
        "os": platform.system(),
        "architecture": platform.machine(),
        "python_major_version": str(sys.version_info.major),
        "python_minor_version": str(sys.version_info.minor),
    }



import prometheus_client

_INF = float("inf")


def latency_metric_for_fast_operation(
    operation_name: str, operation_description: str
) -> prometheus_client.Histogram:
    """Creates a histogram metric for latency of a fast operation.

    A fast operation is typically expected to complete in a few tens of milliseconds."""
    return prometheus_client.Histogram(
        f"{operation_name}_latency_seconds",
        f"Latency of {operation_description}",
        # Buckets are in seconds
        buckets=[
            _ms_to_sec(1),
            _ms_to_sec(2),
            _ms_to_sec(3),
            _ms_to_sec(4),
            _ms_to_sec(5),
            _ms_to_sec(6),
            _ms_to_sec(7),
            _ms_to_sec(8),
            _ms_to_sec(9),
            _ms_to_sec(10),
            _ms_to_sec(15),
            _ms_to_sec(20),
            _ms_to_sec(25),
            _ms_to_sec(30),
            _ms_to_sec(35),
            _ms_to_sec(40),
            _ms_to_sec(45),
            _ms_to_sec(50),
            _ms_to_sec(55),
            _ms_to_sec(60),
            _ms_to_sec(65),
            _ms_to_sec(70),
            _ms_to_sec(75),
            _ms_to_sec(80),
            _ms_to_sec(85),
            _ms_to_sec(90),
            _ms_to_sec(95),
            _ms_to_sec(100),
            _ms_to_sec(150),
            _ms_to_sec(200),
            _ms_to_sec(250),
            _ms_to_sec(300),
            _ms_to_sec(350),
            _ms_to_sec(400),
            _ms_to_sec(450),
            _ms_to_sec(500),
            _ms_to_sec(550),
            _ms_to_sec(600),
            _ms_to_sec(650),
            _ms_to_sec(700),
            _ms_to_sec(750),
            _ms_to_sec(800),
            _ms_to_sec(850),
            _ms_to_sec(900),
            _ms_to_sec(950),
            1,
            _ms_to_sec(1050),
            _ms_to_sec(1100),
            _ms_to_sec(1150),
            _ms_to_sec(1200),
            _ms_to_sec(1250),
            _ms_to_sec(1300),
            _ms_to_sec(1350),
            _ms_to_sec(1400),
            _ms_to_sec(1450),
            _ms_to_sec(1500),
            _ms_to_sec(1550),
            _ms_to_sec(1600),
            _ms_to_sec(1650),
            _ms_to_sec(1700),
            _ms_to_sec(1750),
            _ms_to_sec(1800),
            _ms_to_sec(1850),
            _ms_to_sec(1900),
            _ms_to_sec(1950),
            2,
            _ms_to_sec(2100),
            _ms_to_sec(2200),
            _ms_to_sec(2300),
            _ms_to_sec(2400),
            _ms_to_sec(2500),
            _ms_to_sec(2600),
            _ms_to_sec(2700),
            _ms_to_sec(2800),
            _ms_to_sec(2900),
            3,
            _ms_to_sec(3500),
            4,
            _ms_to_sec(4500),
            5,
            10,
            60,
            _INF,
        ],
    )


def latency_metric_for_slow_operation(
    operation_name: str, operation_description: str
) -> prometheus_client.Histogram:
    """Creates a histogram metric for latency of a slow operation.

    A slow operation is typically expected to complete within a few tens of seconds."""
    return prometheus_client.Histogram(
        f"{operation_name}_latency_seconds",
        f"Latency of {operation_description}",
        # Buckets are in seconds
        buckets=[
            _ms_to_sec(10),
            _ms_to_sec(20),
            _ms_to_sec(30),
            _ms_to_sec(40),
            _ms_to_sec(50),
            _ms_to_sec(60),
            _ms_to_sec(70),
            _ms_to_sec(80),
            _ms_to_sec(90),
            _ms_to_sec(100),
            _ms_to_sec(200),
            _ms_to_sec(300),
            _ms_to_sec(400),
            _ms_to_sec(500),
            _ms_to_sec(600),
            _ms_to_sec(700),
            _ms_to_sec(800),
            _ms_to_sec(900),
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            15,
            20,
            25,
            30,
            35,
            40,
            45,
            50,
            _minutes_to_sec(1),
            _minutes_to_sec(2),
            _minutes_to_sec(3),
            _minutes_to_sec(4),
            _minutes_to_sec(5),
            _minutes_to_sec(6),
            _minutes_to_sec(7),
            _minutes_to_sec(8),
            _minutes_to_sec(9),
            _minutes_to_sec(10),
            _minutes_to_sec(20),
            _INF,
        ],
    )


def latency_metric_for_customer_controlled_operation(
    operation_name: str, operation_description: str
) -> prometheus_client.Histogram:
    """Creates a histogram metric for latency of a customer controlled operation.

    Example of a customer controlled operation is executing customer code.
    The buckets in this histrogram tend to be quite large because we don't
    know how long customer code will take to execute and because we don't need
    a precise understanding of its duration."""
    return prometheus_client.Histogram(
        f"{operation_name}_latency_seconds",
        f"Latency of {operation_description}",
        # Buckets are in seconds
        buckets=[
            _ms_to_sec(1),
            _ms_to_sec(5),
            _ms_to_sec(50),
            _ms_to_sec(100),
            _ms_to_sec(200),
            _ms_to_sec(300),
            _ms_to_sec(400),
            _ms_to_sec(500),
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            15,
            20,
            30,
            40,
            50,
            _minutes_to_sec(1),
            _minutes_to_sec(2),
            _minutes_to_sec(3),
            _minutes_to_sec(4),
            _minutes_to_sec(5),
            _minutes_to_sec(6),
            _minutes_to_sec(7),
            _minutes_to_sec(8),
            _minutes_to_sec(9),
            _minutes_to_sec(10),
            _minutes_to_sec(20),
            _minutes_to_sec(30),
            _minutes_to_sec(40),
            _minutes_to_sec(50),
            _hours_to_sec(1),
            _hours_to_sec(2),
            _hours_to_sec(3),
            _hours_to_sec(4),
            _hours_to_sec(5),
            _hours_to_sec(10),
            _hours_to_sec(20),
            _days_to_sec(1),
            _days_to_sec(2),
            _INF,
        ],
    )


def _ms_to_sec(ms: int) -> float:
    return ms / 1000.0


def _minutes_to_sec(minutes: int) -> float:
    return minutes * 60.0


def _hours_to_sec(hours: int) -> float:
    return _minutes_to_sec(hours * 60)


def _days_to_sec(days: int) -> float:
    return _hours_to_sec(days * 24)


class IdempotentCounterChanger:
    """A wrapper that ensures that inc/dec operations on a counter are done only once.

    This is useful for tracking the number of in-progress operations or objects that exist.
    """

    def __init__(self, counter: prometheus_client.Counter) -> None:
        self._counter: prometheus_client.Counter = counter
        self.is_incremented: bool = False
        self.is_decremented: bool = False

    def inc(self) -> None:
        if not self.is_incremented:
            self._counter.inc()
            self.is_incremented = True

    def dec(self) -> None:
        if self.is_incremented and not self.is_decremented:
            self._counter.dec()
            self.is_decremented = True



import asyncio
import os
from collections.abc import Awaitable, Callable
from typing import Any

import grpc
import grpc.aio
from tensorlake.function_executor.proto.function_executor_pb2 import (
    HealthCheckRequest,
    HealthCheckResponse,
)
from tensorlake.function_executor.proto.function_executor_pb2_grpc import (
    FunctionExecutorStub,
)

from .metrics.health_checker import (
    metric_failed_health_checks,
    metric_health_check_latency,
)
from .server.client_configuration import HEALTH_CHECK_TIMEOUT_SEC

# Use lowest feasible value for now to detect FE crashes quickly because
# we're only doing periodic health checks now.
HEALTH_CHECK_POLL_PERIOD_SEC = 5


class HealthCheckResult:
    def __init__(self, is_healthy: bool, reason: str):
        self.is_healthy: bool = is_healthy
        self.reason: str = reason


class HealthChecker:
    def __init__(
        self, channel: grpc.aio.Channel, stub: FunctionExecutorStub, logger: Any
    ):
        self._channel: grpc.aio.Channel = channel
        self._stub: FunctionExecutorStub = stub
        self._logger: Any = logger.bind(module=__name__)
        self._health_check_loop_task: asyncio.Task | None = None
        self._health_check_failed_callback: Callable[
            [HealthCheckResult], Awaitable[None] | None
        ] = None

    async def check(self) -> HealthCheckResult:
        """Runs the health check once and returns the result.

        Does not raise any exceptions."""
        if os.getenv("INDEXIFY_DISABLE_FUNCTION_EXECUTOR_HEALTH_CHECKS", "0") == "1":
            return HealthCheckResult(
                is_healthy=True,
                reason="Function Executor health checks are disabled using INDEXIFY_DISABLE_FUNCTION_EXECUTOR_HEALTH_CHECKS env var.",
            )

        with metric_health_check_latency.time():
            try:
                response: HealthCheckResponse = await self._stub.check_health(
                    HealthCheckRequest(), timeout=HEALTH_CHECK_TIMEOUT_SEC
                )
                if not response.healthy:
                    metric_failed_health_checks.inc()
                return HealthCheckResult(
                    is_healthy=response.healthy, reason=response.status_message
                )
            except grpc.aio.AioRpcError as e:
                # Due to the customer code running in Function Executor we can't reliably conclude
                # that the FE is unhealthy when RPC status code is not OK. E.g. customer code can
                # hold Python GIL and prevent the health check RPC from being processed by FE Python code.
                #
                # The only unhealthy condition we can be sure about is when the channel can't re-establish
                # the TCP connection within HEALTH_CHECK_TIMEOUT_SEC deadline. This is because FE Python
                # code is not involved when TCP connections are established to FE. Problems reestablishing
                # the TCP connection are usually due to the FE process crashing and its gRPC server socket
                # not being available anymore or due to prolonged local networking failures on Executor.
                if (
                    _channel_state(self._channel, self._logger)
                    == grpc.ChannelConnectivity.TRANSIENT_FAILURE
                ):
                    return HealthCheckResult(
                        is_healthy=False,
                        reason="Channel is in TRANSIENT_FAILURE state, assuming Function Executor crashed.",
                    )
                else:
                    return HealthCheckResult(
                        is_healthy=True,
                        reason=f"Health check RPC failed with status code: {e.code().name}. Assuming Function Executor is healthy.",
                    )
            except Exception as e:
                self._logger.error("Got unexpected exception, ignoring", exc_info=e)
                return HealthCheckResult(
                    is_healthy=True,
                    reason=f"Unexpected exception in Executor: {str(e)}. Assuming Function Executor is healthy.",
                )

    def start(self, callback: Callable[[HealthCheckResult], Awaitable[None]]) -> None:
        """Starts periodic health checks.

        The supplied callback is an async function called in the calling thread's
        event loop when the health check fails. The callback is called only once
        and then health checks are stopped.

        Without a periodic health check a TCP client socket won't detect a server
        socket problem (e.g. it's closed due to server crash) because there are no
        TCP packets floating between them.

        Does not raise any exceptions.
        """
        if self._health_check_loop_task is not None:
            return

        self._health_check_failed_callback = callback
        self._health_check_loop_task = asyncio.create_task(
            self._health_check_loop(), name="function executor health checker loop"
        )

    def stop(self) -> None:
        """Stops the periodic health checks.

        Does not raise any exceptions."""
        if self._health_check_loop_task is None:
            return

        self._health_check_loop_task.cancel()
        self._health_check_loop_task = None

    async def _health_check_loop(self) -> None:
        while True:
            result: HealthCheckResult = await self.check()
            if not result.is_healthy:
                break
            await asyncio.sleep(HEALTH_CHECK_POLL_PERIOD_SEC)

        asyncio.create_task(
            self._health_check_failed_callback(result),
            name="function executor health check failure callback",
        )
        self._health_check_loop_task = None


def _channel_state(channel: grpc.aio.Channel, logger: Any) -> grpc.ChannelConnectivity:
    """Get channel connectivity state and suppresses all exceptions.

    Suppressing the exceptions is important because the channel connectivity state is an experimental
    feature. On error fallse back to READY state which assumes that the channel is okay.
    """
    try:
        return channel.get_state()
    except Exception as e:
        logger.error(
            "Failed getting channel state, falling back to default READY state",
            exc_info=e,
        )
        return grpc.ChannelConnectivity.READY



import importlib
import os
import traceback
from typing import Any, Generator

import click
import docker
import docker.api.build
from docker.errors import BuildError
from docker.models.images import Image as DockerImage
from tensorlake.applications import Image
from tensorlake.applications.image import (
    ImageInformation,
    dockerfile_content,
    image_infos,
)
from tensorlake.applications.remote.code.loader import load_code


@click.command(
    short_help="Builds images for applications defined in <application-file-path> .py file"
)
@click.argument(
    "application-file-path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
def build_images(application_file_path: str):
    try:
        application_file_path: str = os.path.abspath(application_file_path)
        load_code(application_file_path)
    except Exception as e:
        click.secho(
            f"Failed to load the code directory modules, please check the error message: {e}",
            fg="red",
        )
        traceback.print_exception(e)
        raise click.Abort

    docker_client: docker.DockerClient = docker.from_env()
    docker_client.ping()

    indexify_version: str = importlib.metadata.version("indexify")
    for image, image_info in image_infos().items():
        image: Image
        image_info: ImageInformation
        click.echo(f"Building image `{image.name}:{image.tag}`")
        image.run(f"pip install 'indexify=={indexify_version}'")

        try:
            built_image, logs_generator = _build(
                image=image, docker_client=docker_client
            )
            built_image: DockerImage
            _print_build_log(logs_generator)
            click.secho(f"Built image: {built_image.tags[0]}", fg="green")
        except BuildError as e:
            traceback.print_exception(e)
            raise click.Abort


def _build(
    image: Image, docker_client: docker.DockerClient
) -> tuple[DockerImage, Generator[str, Any, None]]:
    docker_file_content: str = dockerfile_content(image)
    image_name = image.name if ":" in image.name else f"{image.name}:{image.tag}"

    docker.api.build.process_dockerfile = lambda dockerfile, path: (
        "Dockerfile",
        dockerfile,
    )

    try:
        built_image, logs_generator = docker_client.images.build(
            path=".",
            dockerfile=docker_file_content,
            tag=image_name,
            rm=True,
            # pull=True,  # optional: ensures fresh base images
            # forcerm=True,  # optional: always remove intermediate containers
        )
        return built_image, logs_generator
    except BuildError as e:
        click.secho("Docker build failed:", fg="red")
        _print_build_log(e.build_log or [])
        click.secho(str(e), fg="red")
        raise


def _print_build_log(build_logs):
    for log_entry in build_logs:
        if isinstance(log_entry, dict):
            if "stream" in log_entry:
                click.echo(log_entry["stream"].rstrip("\n"))
            elif "status" in log_entry:
                if "id" in log_entry:
                    click.echo(f"{log_entry['status']}: {log_entry['id']}")
                else:
                    click.echo(log_entry["status"])
            if "errorDetail" in log_entry:
                # This is the most useful bit when a RUN command fails
                msg = log_entry["errorDetail"].get("message") or log_entry.get("error")
                if msg:
                    click.secho(msg.rstrip("\n"), fg="red")
        elif isinstance(log_entry, str):
            click.echo(log_entry.rstrip("\n"))



# Executor logging configuration.
import logging
import sys
import traceback
from typing import TextIO

import structlog


def configure_logging_early():
    """Configures standard Python logging module.
    By default 3rd party modules that are using standard Python logging module
    (logging.getLogger()) have their log lines dropped unless the module gets configured.
    Not dropping log lines from 3rd party modules is useful for debugging. E.g. this helps
    debugging errors in grpc servers if exceptions happen inside grpc module code.
    """
    logging.basicConfig(
        level=logging.WARNING,
        # This log message format is a bit similar to the default structlog format.
        format="%(asctime)s [%(levelname)s] %(message)s logger=%(name)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )


def configure_development_mode_logging(compact_tracebacks=True):
    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
        structlog.dev.set_exc_info,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S", utc=False),
        structlog.dev.ConsoleRenderer(
            exception_formatter=(
                _compact_traceback_formatter
                if compact_tracebacks
                else structlog.dev.rich_traceback
            ),
        ),
    ]
    structlog.configure(
        processors=processors,
    )


def configure_production_mode_logging():
    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
        structlog.dev.set_exc_info,
        structlog.processors.format_exc_info,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S", utc=False),
        structlog.processors.JSONRenderer(),
    ]
    structlog.configure(processors=processors)


def _compact_traceback_formatter(
    sio: TextIO, exc_info: structlog.typing.ExcInfo
) -> None:
    print(
        "\n (only the frame where the exception was raised is printed by default)",
        file=sio,
    )
    traceback.print_exception(
        *exc_info,
        file=sio,
        limit=-1,  # Print only 1 frame where the exception was raised
        chain=False,
    )



import click

from . import build_images, deploy, executor


@click.group()
@click.version_option(package_name="indexify", prog_name="indexify-cli")
@click.pass_context
def cli(ctx: click.Context):
    """
    Indexify CLI to manage and deploy applications to Indexify Server and run Indexify Executors.
    """
    pass


cli.add_command(build_images.build_images)
cli.add_command(deploy.deploy)
cli.add_command(executor.executor)



import asyncio
from typing import Any, List, Optional

from indexify.executor.function_executor.function_executor import FunctionExecutor
from indexify.proto.executor_api_pb2 import FunctionExecutorTerminationReason

from .aio_utils import shielded_await
from .events import FunctionExecutorTerminated


async def terminate_function_executor(
    function_executor: Optional[FunctionExecutor],
    lock: asyncio.Lock,
    fe_termination_reason: FunctionExecutorTerminationReason,
    allocation_ids_caused_termination: List[str],
    logger: Any,
) -> FunctionExecutorTerminated:
    """Destroys the function executor if it's not None.

    The supplied lock is used to ensure that if a destroy operation is in progress,
    then another caller won't return immediately assuming that the destroy is complete
    due to its idempotency. Ignores cancellations while destroying the function executor.

    Doesn't raise any exceptions.
    """
    logger = logger.bind(module=__name__)

    if function_executor is not None:
        async with lock:
            logger.info(
                "destroying function executor",
            )
            fe_destroy_task: asyncio.Task = asyncio.create_task(
                function_executor.destroy(),
                name="destroy function executor",
            )
            try:
                await shielded_await(fe_destroy_task, logger)
            except asyncio.CancelledError:
                # We actually destroyed the FE so we can return without error.
                pass

    return FunctionExecutorTerminated(
        is_success=True,
        fe_termination_reason=fe_termination_reason,
        allocation_ids_caused_termination=allocation_ids_caused_termination,
    )



import asyncio
from pathlib import Path
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Set,
)

from tensorlake.function_executor.proto.message_validator import MessageValidator

from indexify.proto.executor_api_pb2 import (
    Allocation,
    DesiredExecutorState,
    FunctionExecutorDescription,
    GetDesiredExecutorStatesRequest,
)
from indexify.proto.executor_api_pb2_grpc import ExecutorAPIStub

from .blob_store.blob_store import BLOBStore
from .channel_manager import ChannelManager
from .function_executor.server.function_executor_server_factory import (
    FunctionExecutorServerFactory,
)
from .function_executor_controller import (
    FunctionExecutorController,
    allocation_logger,
    function_executor_logger,
    validate_allocation,
    validate_function_executor_description,
)
from .metrics.state_reconciler import (
    metric_state_reconciliation_errors,
    metric_state_reconciliation_latency,
    metric_state_reconciliations,
)
from .state_reporter import ExecutorStateReporter

_RECONCILE_STREAM_BACKOFF_INTERVAL_SEC = 5
_RECONCILIATION_RETRIES = 3
# If we didn't get a new desired state from the stream within this timeout then the stream might
# not be healthy due to network disruption. In this case we need to recreate the stream to make
# sure that Server really doesn't want to send us a new state.
_DESIRED_EXECUTOR_STATES_TIMEOUT_SEC = 5 * 60  # 5 minutes


class ExecutorStateReconciler:
    def __init__(
        self,
        executor_id: str,
        function_executor_server_factory: FunctionExecutorServerFactory,
        base_url: str,
        config_path: str | None,
        cache_path: Path,
        blob_store: BLOBStore,
        channel_manager: ChannelManager,
        state_reporter: ExecutorStateReporter,
        logger: Any,
        server_backoff_interval_sec: int = _RECONCILE_STREAM_BACKOFF_INTERVAL_SEC,
    ):
        self._executor_id: str = executor_id
        self._function_executor_server_factory: FunctionExecutorServerFactory = (
            function_executor_server_factory
        )
        self._base_url: str = base_url
        self._config_path: str | None = config_path
        self._cache_path: Path = cache_path
        self._blob_store: BLOBStore = blob_store
        self._channel_manager: ChannelManager = channel_manager
        self._state_reporter: ExecutorStateReporter = state_reporter
        self._logger: Any = logger.bind(module=__name__)
        self._server_backoff_interval_sec: int = server_backoff_interval_sec

        # Mutable state. Doesn't need lock because we access from async tasks running in the same thread.
        self._desired_states_reader: asyncio.Task | None = None
        self._reconciliation_loop_runner: asyncio.Task | None = None
        self._function_executor_controllers: Dict[str, FunctionExecutorController] = {}
        self._shutting_down_fe_ids: Set[str] = set()
        self._last_server_clock: int | None = None

        self._last_desired_state_lock = asyncio.Lock()
        self._last_desired_state_change_notifier: asyncio.Condition = asyncio.Condition(
            lock=self._last_desired_state_lock
        )
        self._last_desired_state: DesiredExecutorState | None = None

    def get_desired_state(self) -> DesiredExecutorState | None:
        return self._last_desired_state

    def run(self):
        """Runs the state reconciler.

        Never raises any exceptions. Doesn't block.
        """
        if self._reconciliation_loop_runner is not None:
            self._logger.error(
                "reconciliation loop aio task is already running, skipping run call"
            )
            return

        self._reconciliation_loop_runner = asyncio.create_task(
            self._reconciliation_loop(),
            name="state reconciler reconciliation loop",
        )
        self._desired_states_reader = asyncio.create_task(
            self._desired_states_reader_loop(),
            name="state reconciler desired states stream reader",
        )

    async def shutdown(self):
        """Shuts down the state reconciler.

        Never raises any exceptions.
        """
        if self._reconciliation_loop_runner is not None:
            self._reconciliation_loop_runner.cancel()
            try:
                await self._reconciliation_loop_runner
            except asyncio.CancelledError:
                # Expected cancellation, nothing to do.
                pass
            self._logger.info("reconciliation loop is shutdown")

        if self._desired_states_reader is not None:
            self._desired_states_reader.cancel()
            try:
                await self._desired_states_reader
            except asyncio.CancelledError:
                # Expected cancellation, nothing to do.
                pass
            self._logger.info("desired states stream reader loop is shutdown")

        # Now all the aio tasks exited so nothing will intervene with our actions from this point.
        fe_shutdown_aio_tasks: List[asyncio.Task] = []
        for fe_controller in self._function_executor_controllers.values():
            fe_shutdown_aio_tasks.append(
                asyncio.create_task(
                    fe_controller.shutdown(),
                    name=f"Shutdown Function Executor {fe_controller.function_executor_id()}",
                )
            )

        # Run all the shutdown tasks concurrently and wait for them to complete.
        for aio_task in fe_shutdown_aio_tasks:
            await aio_task

        self._function_executor_controllers.clear()
        self._logger.info("state reconciler is shutdown")

    async def _desired_states_reader_loop(self):
        """Reads the desired states stream from Server and processes it.

        Never raises any exceptions. Gets cancelled via aio task cancellation.
        """
        while True:
            desired_states_stream: AsyncIterable[DesiredExecutorState] | None = None
            try:
                stub = ExecutorAPIStub(await self._channel_manager.get_shared_channel())
                # Report state once before starting the stream so Server
                # doesn't use stale state it knew about this Executor in the past.
                await self._state_reporter.report_state_and_wait_for_completion()

                desired_states_stream = stub.get_desired_executor_states(
                    GetDesiredExecutorStatesRequest(executor_id=self._executor_id)
                )
                self._logger.info("created new desired states stream")
                await self._process_desired_states_stream(desired_states_stream)
            except Exception as e:
                self._logger.error(
                    f"error while processing desired states stream",
                    exc_info=e,
                )
            finally:
                # Cleanly signal Server that the stream is closed by client.
                # See https://stackoverflow.com/questions/72207914/how-to-stop-listening-on-a-stream-in-python-grpc-client
                if desired_states_stream is not None:
                    desired_states_stream.cancel()

            self._logger.info(
                f"desired states stream closed, reconnecting in {self._server_backoff_interval_sec} sec"
            )
            await asyncio.sleep(self._server_backoff_interval_sec)

    async def _process_desired_states_stream(
        self, desired_states: AsyncIterable[DesiredExecutorState]
    ):
        desired_states_iter: AsyncIterator[DesiredExecutorState] = aiter(desired_states)
        while True:
            try:
                new_state: DesiredExecutorState = await asyncio.wait_for(
                    anext(desired_states_iter),
                    timeout=_DESIRED_EXECUTOR_STATES_TIMEOUT_SEC,
                )
            except asyncio.TimeoutError:
                # These log lines really help to debug networking issues. When there are
                # no networking issues and the fleet is not idle we don't get excessive logging here.
                self._logger.info(
                    f"No desired state received from Server within {_DESIRED_EXECUTOR_STATES_TIMEOUT_SEC} sec, recreating the stream to ensure it is healthy"
                )
                break  # Timeout reached, stream might be unhealthy, exit the loop to recreate the stream.

            validator: MessageValidator = MessageValidator(new_state)
            try:
                validator.required_field("clock")
            except ValueError as e:
                self._logger.error(
                    "received invalid DesiredExecutorState from Server, ignoring",
                    exc_info=e,
                )
                continue

            # TODO: The clock is only incremented when function executors have actionable changes and not on new allocations.
            #       Therefore the clock cannot currently be used as an idempotency token.
            # if self._last_server_clock is not None:
            #     if self._last_server_clock >= new_state.clock:
            #         self._logger.warning(
            #             "received outdated DesiredExecutorState from Server, ignoring",
            #             current_clock=self._last_server_clock,
            #             ignored_clock=new_state.clock,
            #         )
            #         continue  # Duplicate or outdated message state sent by Server.

            self._last_server_clock = new_state.clock
            # Always read the latest desired state value from the stream so
            # we're never acting on stale desired states.
            async with self._last_desired_state_lock:
                self._last_desired_state = new_state
                self._last_desired_state_change_notifier.notify_all()

    async def _reconciliation_loop(self):
        """Continuously reconciles the desired state with the current state.

        Never raises any exceptions. Get cancelled via aio task cancellation."""
        last_reconciled_state: DesiredExecutorState | None = None
        while True:
            async with self._last_desired_state_lock:
                # Comparing object identities (references) is enough here to not reconcile
                # the same state twice.
                while self._last_desired_state is last_reconciled_state:
                    await self._last_desired_state_change_notifier.wait()
                last_reconciled_state = self._last_desired_state

            with metric_state_reconciliation_latency.time():
                metric_state_reconciliations.inc()
                await self._reconcile_state(last_reconciled_state)
                # Update the clock regardless of success or failure.
                # This is to show Server that we actually processed the message.
                self._state_reporter.update_last_server_clock(
                    last_reconciled_state.clock
                )

    async def _reconcile_state(self, desired_state: DesiredExecutorState):
        """Reconciles the desired state with the current state.

        Doesn't raise any exceptions. Logs all errors for future investigation becase the gRPC protocol
        doesn't allow us to return errors to the Server if it supplied invalid messages.
        """
        for attempt in range(_RECONCILIATION_RETRIES):
            try:
                # Reconcile FEs first because allocations depend on them.
                self._reconcile_function_executors(desired_state.function_executors)
                self._reconcile_allocations(desired_state.allocations)
                return
            except Exception as e:
                self._logger.error(
                    "failed to reconcile desired state, retrying in 5 secs",
                    exc_info=e,
                    attempt=attempt,
                    attempts_left=_RECONCILIATION_RETRIES - attempt,
                )
                await asyncio.sleep(5)

        metric_state_reconciliation_errors.inc()
        self._logger.error(
            f"failed to reconcile desired state after {_RECONCILIATION_RETRIES} attempts",
        )

    def _reconcile_function_executors(
        self, function_executor_descriptions: Iterable[FunctionExecutorDescription]
    ):
        valid_fe_descriptions: List[FunctionExecutorDescription] = (
            self._valid_function_executor_descriptions(function_executor_descriptions)
        )
        for fe_description in valid_fe_descriptions:
            self._reconcile_function_executor(fe_description)

        seen_fe_ids: Set[str] = set(map(lambda fe: fe.id, valid_fe_descriptions))
        fe_ids_to_remove = set(self._function_executor_controllers.keys()) - seen_fe_ids
        for fe_id in fe_ids_to_remove:
            # Server forgot this FE, so its safe to forget about it now too.
            self._remove_function_executor_controller(fe_id)

    def _valid_function_executor_descriptions(
        self, function_executor_descriptions: Iterable[FunctionExecutorDescription]
    ):
        valid_function_executor_descriptions: List[FunctionExecutorDescription] = []
        for function_executor_description in function_executor_descriptions:
            function_executor_description: FunctionExecutorDescription
            logger = function_executor_logger(
                function_executor_description, self._logger
            )

            try:
                validate_function_executor_description(function_executor_description)
            except ValueError as e:
                logger.error(
                    "received invalid FunctionExecutorDescription from Server, dropping it from desired state",
                    exc_info=e,
                )
                continue

            valid_function_executor_descriptions.append(function_executor_description)

        return valid_function_executor_descriptions

    def _reconcile_function_executor(
        self, function_executor_description: FunctionExecutorDescription
    ):
        """Reconciles a single Function Executor with the desired state.

        Doesn't block on any long running operations. Doesn't raise any exceptions.
        """

        if function_executor_description.id not in self._function_executor_controllers:
            self._add_function_executor_controller(function_executor_description)

    def _add_function_executor_controller(
        self, function_executor_description: FunctionExecutorDescription
    ) -> None:
        """Creates Function Executor for the supplied description and adds it to internal data structures.

        Doesn't block on any long running operations. Doesn't raise any exceptions.
        """
        logger = function_executor_logger(function_executor_description, self._logger)
        try:
            controller: FunctionExecutorController = FunctionExecutorController(
                executor_id=self._executor_id,
                function_executor_description=function_executor_description,
                function_executor_server_factory=self._function_executor_server_factory,
                state_reporter=self._state_reporter,
                blob_store=self._blob_store,
                base_url=self._base_url,
                config_path=self._config_path,
                cache_path=self._cache_path,
                logger=self._logger,
            )
            self._function_executor_controllers[function_executor_description.id] = (
                controller
            )
            controller.startup()
        except Exception as e:
            logger.error("failed adding Function Executor", exc_info=e)

    def _remove_function_executor_controller(self, function_executor_id: str) -> None:
        # Don't remove the FE controller from self._function_executor_controllers until
        # its shutdown is complete. Otherwise, if Server re-adds the FE to desired state
        # before FE shutdown completes then we'll have two FE controllers for the same
        # FE ID which results in many bugs.
        if function_executor_id in self._shutting_down_fe_ids:
            return

        self._shutting_down_fe_ids.add(function_executor_id)
        asyncio.create_task(
            self._shutdown_function_executor_controller(function_executor_id),
            name=f"Shutdown Function Executor {function_executor_id}",
        )

    async def _shutdown_function_executor_controller(
        self, function_executor_id: str
    ) -> None:
        # We are not cancelling this aio task in self.shutdown(). Because of this the code here should
        # not fail if the FE controller is not found in internal data structures. It can be removed
        # by self.shutdown() at any time while we're running this aio task.
        fe_controller: FunctionExecutorController | None = (
            self._function_executor_controllers.get(function_executor_id)
        )
        if fe_controller is None:
            return

        await fe_controller.shutdown()
        self._function_executor_controllers.pop(function_executor_id, None)
        self._shutting_down_fe_ids.discard(function_executor_id)

    def _reconcile_allocations(self, allocations: Iterable[Allocation]):
        valid_allocations: List[Allocation] = self._valid_allocations(allocations)
        for allocation in valid_allocations:
            self._reconcile_allocation(allocation)

        # Cancel allocs that are no longer in the desired state.
        # FE ID => [Allocation ID]
        desired_alloc_ids_per_fe: Dict[str, List[str]] = {}
        for allocation in valid_allocations:
            if allocation.function_executor_id not in desired_alloc_ids_per_fe:
                desired_alloc_ids_per_fe[allocation.function_executor_id] = []
            desired_alloc_ids_per_fe[allocation.function_executor_id].append(
                allocation.allocation_id
            )

        for fe_controller in self._function_executor_controllers.values():
            fe_controller: FunctionExecutorController
            if fe_controller.function_executor_id() in desired_alloc_ids_per_fe:
                desired_fe_alloc_ids: Set[str] = set(
                    desired_alloc_ids_per_fe[fe_controller.function_executor_id()]
                )
            else:
                # No allocations desired for this FE, so cancel all its allocations.
                desired_fe_alloc_ids: Set[str] = set()
            actual_fe_alloc_ids: Set[str] = set(fe_controller.allocation_ids())
            alloc_ids_to_remove: Set[str] = actual_fe_alloc_ids - desired_fe_alloc_ids
            for alloc_id in alloc_ids_to_remove:
                fe_controller.remove_allocation(alloc_id)

    def _reconcile_allocation(self, allocation: Allocation):
        """Reconciles a single Allocation with the desired state.

        Doesn't raise any exceptions.
        """
        function_executor_controller: FunctionExecutorController = (
            self._function_executor_controllers[allocation.function_executor_id]
        )
        if function_executor_controller.has_allocation(allocation.allocation_id):
            # Nothing to do, allocation already exists and it's immutable.
            return

        function_executor_controller.add_allocation(allocation)

    def _valid_allocations(self, allocations: Iterable[Allocation]):
        valid_allocations: List[Allocation] = []
        for allocation in allocations:
            allocation: Allocation
            logger = allocation_logger(allocation, self._logger)

            try:
                validate_allocation(allocation)
            except ValueError as e:
                # There's no way to report this error to Server so just log it.
                logger.error(
                    "received invalid Allocation from Server, dropping it from desired state",
                    exc_info=e,
                )
                continue

            if (
                allocation.function_executor_id
                not in self._function_executor_controllers
            ):
                logger.error(
                    "received Allocation for a Function Executor that doesn't exist, dropping it from desired state"
                )
                continue

            valid_allocations.append(allocation)

        return valid_allocations



import prometheus_client

from indexify.executor.monitoring.metrics import (
    latency_metric_for_customer_controlled_operation,
)

metric_function_executor_run_allocation_rpcs: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_run_allocation_rpcs",
        "Number of Function Executor run allocation lifecycle RPC sequences",
    )
)
metric_function_executor_run_allocation_rpc_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_run_allocation_rpc_errors",
        "Number of Function Executor run allocation lifecycle RPC errors",
    )
)
metric_function_executor_run_allocation_rpc_latency: prometheus_client.Histogram = (
    latency_metric_for_customer_controlled_operation(
        "function_executor_run_allocation_rpc",
        "Function Executor run allocation lifecycle RPC",
    )
)
metric_function_executor_run_allocation_rpcs_in_progress: prometheus_client.Gauge = (
    prometheus_client.Gauge(
        "function_executor_run_allocation_rpcs_in_progress",
        "Number of Function Executor run allocation lifecycle RPCs in progress",
    )
)



import asyncio
from typing import Any, AsyncGenerator, Union

import grpc
import httpx
from tensorlake.function_executor.proto.function_executor_pb2 import (
    GetRequestStateResponse,
    RequestStateRequest,
    RequestStateResponse,
    SerializedObject,
    SerializedObjectEncoding,
    SerializedObjectManifest,
    SetRequestStateResponse,
)
from tensorlake.function_executor.proto.function_executor_pb2_grpc import (
    FunctionExecutorStub,
)
from tensorlake.function_executor.proto.message_validator import MessageValidator

from .metrics.request_state_client import (
    metric_request_read_errors,
    metric_server_get_state_request_errors,
    metric_server_get_state_request_latency,
    metric_server_get_state_requests,
    metric_server_set_state_request_errors,
    metric_server_set_state_request_latency,
    metric_server_set_state_requests,
)

# We're currently only supporting Pickle for request state values.
# FIXME: if Function Executor sends us something else then we fail the calls.
_VALUE_CONTENT_TYPE = "application/octet-stream"


class RequestStateClient:
    """RequestStateClient is a client for the request state server of a Function Executor.

    The client initializes the Function Executor's request state server and executes requests
    it sends to the client.
    """

    def __init__(
        self,
        stub: FunctionExecutorStub,
        base_url: str,
        http_client: httpx.AsyncClient,
        application: str,
        namespace: str,
        logger: Any,
    ):
        self._stub: FunctionExecutorStub = stub
        self._base_url: str = base_url
        self._http_client: httpx.AsyncClient = http_client
        self._application_name: str = application
        self._namespace: str = namespace
        self._logger: Any = logger.bind(
            module=__name__, application=application, namespace=namespace
        )
        self._client_response_queue: asyncio.Queue[Union[RequestStateResponse, str]] = (
            asyncio.Queue()
        )
        self._allocation_id_to_request_id: dict[str, str] = {}
        self._request_loop_task: asyncio.Task | None = None

    async def start(self) -> None:
        """Starts the request state client.

        This method initializes the Function Executor's request state server first.
        This is why this method needs to be awaited before executing any tasks on the Function Executor
        that might use request state feature."""
        server_requests = self._stub.initialize_request_state_server(
            self._response_generator()
        )
        self._request_loop_task = asyncio.create_task(
            self._request_loop(server_requests),
            name="graph request state client request processing loop",
        )

    def add_allocation_to_request_id_entry(
        self, allocation_id: str, request_id: str
    ) -> None:
        """Adds a allocation ID to request ID entry to the client's internal state.

        This allows to authorize requests to the request state server.
        If a request is not comming from the allocation ID that was added here then it will
        be rejected. It's caller's responsibility to only add task IDs that are being
        executed by the Function Executor so the Function Executor can't get access to
        request state of tasks it doesn't run.

        Doesn't raise any exceptions.
        """
        self._allocation_id_to_request_id[allocation_id] = request_id

    def remove_allocation_to_request_id_entry(self, allocation_id: str) -> None:
        """Removes a allocation ID to request ID entry from the client's internal state.

        Doesn't raise any exceptions.
        """
        self._allocation_id_to_request_id.pop(allocation_id, None)

    async def destroy(self) -> None:
        if self._request_loop_task is not None:
            self._request_loop_task.cancel()
        await self._client_response_queue.put("shutdown")

    async def _request_loop(
        self, server_requests: AsyncGenerator[RequestStateRequest, None]
    ) -> None:
        try:
            async for request in server_requests:
                await self._process_request_no_raise(request)
        except grpc.aio.AioRpcError:
            # Reading from the stream failed.
            # This is a normal situation when the server is shutting down.
            pass
        except asyncio.CancelledError:
            # This async task was cancelled by destroy(). Normal situation too.
            # This exception should not be suppressed, see Python asyncio docs.
            raise
        except Exception as e:
            metric_request_read_errors.inc()
            self._logger.error(
                "failed to read request from server, shutting down request state client",
                exc_info=e,
            )

    async def _process_request_no_raise(self, request: RequestStateRequest) -> None:
        try:
            await self._process_request(request)
        except Exception as e:
            try:
                await self._client_response_queue.put(
                    RequestStateResponse(
                        state_request_id=request.state_request_id,
                        success=False,
                    )
                )
            except Exception as ee:
                self._logger.error("failed to send error response", exc_info=ee)

            self._logger.error(
                "failed to process request",
                exc_info=e,
                state_request_id=request.state_request_id,
            )

    async def _process_request(
        self, request: RequestStateRequest
    ) -> RequestStateResponse:
        self._validate_request(request)
        # This is a very important check. We don't trust request ID and allocation ID
        # supplied by Function Executor. If a allocation ID entry doesn't exist then it's
        # a privelege escalation attempt.
        request_id: str = self._allocation_id_to_request_id[request.allocation_id]
        if request.HasField("get"):
            with (
                metric_server_get_state_request_errors.count_exceptions(),
                metric_server_get_state_request_latency.time(),
            ):
                metric_server_get_state_requests.inc()
                value: SerializedObject | None = await self._get_server_state(
                    request_id, request.get.key
                )
            await self._client_response_queue.put(
                RequestStateResponse(
                    state_request_id=request.state_request_id,
                    success=True,
                    get=GetRequestStateResponse(
                        key=request.get.key,
                        value=value,
                    ),
                )
            )
        elif request.HasField("set"):
            with (
                metric_server_set_state_request_errors.count_exceptions(),
                metric_server_set_state_request_latency.time(),
            ):
                metric_server_set_state_requests.inc()
                await self._set_server_state(
                    request_id, request.set.key, request.set.value
                )
            await self._client_response_queue.put(
                RequestStateResponse(
                    state_request_id=request.state_request_id,
                    success=True,
                    set=SetRequestStateResponse(),
                )
            )

    async def _response_generator(
        self,
    ) -> AsyncGenerator[RequestStateResponse, None]:
        while True:
            response = await self._client_response_queue.get()
            # Hacky cancellation of the generator.
            if response == "shutdown":
                break
            yield response

    async def _set_server_state(
        self, request_id: str, key: str, value: SerializedObject
    ) -> None:
        url: str = (
            f"{self._base_url}/internal/namespaces/{self._namespace}/applications/{self._application_name}/requests/{request_id}/ctx/{key}"
        )
        if (
            value.manifest.encoding
            != SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_BINARY_PICKLE
        ):
            raise ValueError(
                f"Unsupported value encoding: {SerializedObjectEncoding.Name(value.encoding)}. "
                "Only binary pickle is supported for application request state values."
            )

        response = await self._http_client.post(
            url=url,
            files=[
                (
                    "value",
                    ("value", value.data, _VALUE_CONTENT_TYPE),
                ),
            ],
        )

        try:
            response.raise_for_status()
        except Exception as e:
            self._logger.error(
                "failed to set application request state",
                request_id=request_id,
                key=key,
                status_code=response.status_code,
                error=response.text,
                exc_info=e,
            )
            raise

    async def _get_server_state(
        self, request_id: str, key: str
    ) -> SerializedObject | None:
        url: str = (
            f"{self._base_url}/internal/namespaces/{self._namespace}/applications/{self._application_name}/requests/{request_id}/ctx/{key}"
        )

        response: httpx.Response = await self._http_client.get(url)
        if response.status_code == 404:
            return None

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            self._logger.error(
                f"failed to download application request state value",
                request_id=request_id,
                key=key,
                status_code=response.status_code,
                error=response.text,
                exc_info=e,
            )
            raise

        return _serialized_object_from_http_response(response)

    def _validate_request(self, request: RequestStateRequest) -> None:
        (
            MessageValidator(request)
            .required_field("state_request_id")
            .required_field("allocation_id")
        )
        if request.HasField("get"):
            (MessageValidator(request.get).required_field("key"))
        elif request.HasField("set"):
            (
                MessageValidator(request.set)
                .required_field("key")
                .required_serialized_object("value")
            )
        else:
            raise ValueError("unknown request type")


def _serialized_object_from_http_response(response: httpx.Response) -> SerializedObject:
    if response.headers["content-type"] != _VALUE_CONTENT_TYPE:
        raise ValueError(
            f"Unexpected content type: {response.headers['content-type']}. "
            f"Expected: {_VALUE_CONTENT_TYPE}."
        )

    return SerializedObject(
        manifest=SerializedObjectManifest(
            encoding=SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_BINARY_PICKLE,
            encoding_version=0,
            size=len(response.content),
            metadata_size=0,
            # We don't store any hash on the server side right now and it's not safe
            # to compute it here as this is user manipulated data.
            sha256_hash="fake_hash",
        ),
        data=response.content,
    )



import prometheus_client

from ..monitoring.metrics import latency_metric_for_fast_operation

metric_state_reconciliations = prometheus_client.Counter(
    "state_reconciliations",
    "Number of Executor state reconciliations",
)
metric_state_reconciliation_errors = prometheus_client.Counter(
    "state_reconciliation_errors",
    "Number of Executor state reconciliation errors after all retries",
)
metric_state_reconciliation_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "state_reconciliation", "Executor state reconciliation"
    )
)
